from .command import CmdTypes, CloModes, MrgModes, Cmd, Set, Add, Rem, Clo, Cmp, Get
from .result import Res,ResTypes
from ..query.query import Qry,Obj,ObjSeg,His
from ..util.error import EoqError
from ..util.logger import NoLogging
from ..serialization.jsonserializer import JsonSerializer
from ..serialization.textserializer import TextSerializer

from datetime import time
from uuid import uuid4

from timeit import default_timer as timer

class DifferenceFinder():
    def __init__(self, commandRunner, mdbAccessor, old, new, tid, mrgMode, logger = NoLogging()):
        # properties from commandrunner
        self.commandRunner = commandRunner
        self.mdbAccessor = mdbAccessor
        self.logger = logger
        self.logSerializer = JsonSerializer()
        self.TestSerializer = TextSerializer()
        self.tid = tid
        self.oldRoot = old
        self.newRoot = new
        # set merge mode
        self.writeLog = False
        if mrgMode in [MrgModes.LOG, MrgModes.DET]:
            self.writeLog = True
        # global properties
        self.treatedObjs = [] # for remove loop
        self.clonedObjs = [] # for remove loop and update refs of clones
        self.postponedRefs = [] # for directly after merge routine
        self.refsToUpdate = [] # for update refs
        self.cloneChildren = [] # for children of clones
        self.cloneList = [] # for original clone
        self.modifiedReferences = []
        # properties to be returned to commandrunner
        self.cmdList = []

    # PUBLIC
    def CreateObjectOverviewFile(self):
        ''' Creates a txt file with all objects and their respective #123 numbers
        for this execution of the eoq code. '''
        createOld = True
        createNew = True
        old = []
        new = []
        if createOld:
            self.__SubroutineForCreateObjectOverviewFile(old, self.oldRoot)
            path = 'log/ObjectOverviewOld_{}_{}.txt'.format(self.oldRoot, self.newRoot)
            old.sort()
            with open(path, 'w') as file:
                for entry in old:
                    file.write("{} {}\n".format(entry[0], entry[1]))
        if createNew:
            self.__SubroutineForCreateObjectOverviewFile(new, self.newRoot)
            path = 'log/ObjectOverviewNew_{}_{}.txt'.format(self.oldRoot, self.newRoot)
            new.sort()
            with open(path, 'w') as file:
                for entry in new:
                    file.write("{} {}\n".format(entry[0], entry[1]))
        old.clear()
        new.clear()

    def Run(self):
        '''
        Public function that is called from ExecDif.
        Starts all the sub-functions.
        '''
        self.__StartMergeRoutine()
        self.__StartUpdateReferences()
        self.__StartRemoveLoop()

    def GetCommands(self):
        '''
        Returns all commands necessary to obtain a fully merged object.
        Commands are generated by Run / MergeRoutine and stored in cmdList.
        '''
        res = []
        for cmd in self.cmdList:
            if cmd not in res:
                # correct __UpdateCloneReferences() commands
                if cmd.cmd != 'CLO' and cmd.a[0] in self.clonedObjs:
                    # find corresponding cmp command and append
                    cmpcmd = None
                    for c in res:
                        if c.cmd == 'CMP' and c.a[0].a[0] == cmd.a[0]:
                            cmpcmd = c
                            break
                    # get amount of cmds until original Clone command
                    n = -1*(len(cmpcmd.a))
                    # print('n:', n)
                    if cmd.cmd == 'SET':
                        cmpcmd.Set(His(n), cmd.a[1], cmd.a[2])
                    elif cmd.cmd == 'REM':
                        cmpcmd.Rem(His(n), cmd.a[1], cmd.a[2])
                    elif cmd.cmd == 'ADD':
                        cmpcmd.Add(His(n), cmd.a[1], cmd.a[2])
                # correct commands regarding children of clones
                elif cmd.a[0] in self.cloneChildren:
                    # print('Clone children:', self.cloneChildren)
                    # print('Clone child command:', cmd.cmd, cmd.a)
                    # get original clone
                    originalClone = None
                    for entry in self.cloneList:
                        if entry[0] == cmd.a[0]:
                            originalClone = entry[1]
                            break
                    # print('Original Clone:', originalClone)
                    # add to cmp clone command
                    cmpcmd = None
                    for c in res:
                        if c.cmd == 'CMP' and c.a[0].a[0] == originalClone:
                            cmpcmd = c
                            break
                    # get amount of cmds until original Clone command
                    n = -1*(len(cmpcmd.a))
                    # print('n:', n)
                    # am I a direct child of original clone
                    stammbaum = []
                    if cmd.a[0] not in self.__GetAllChildrenInOneList(originalClone):
                        # we need to create a small stammbaum
                        # indicating all elements between clone child and original clone
                        currentElement = cmd.a[0]
                        while True:
                            parent = self.mdbAccessor.GetParent(currentElement)
                            if parent == originalClone:
                                break
                            stammbaum.append(parent)
                            currentElement = parent
                    # print('Stammbaum:', stammbaum)
                    # get my index
                    idx = self.mdbAccessor.GetIndex(cmd.a[0])
                    # print('My Index:', idx)
                    # get my classname
                    classname = self.__Decode(cmd.a[0]).eClass.name
                    # print('My Classname:', classname)
                    # get my containing feature
                    cf = self.mdbAccessor.GetContainingFeature(cmd.a[0])
                    cfName = self.__Decode(cf).name
                    # print('My containing feature:', cf, cfName)
                    cmdToCloneChild = Qry()
                    cmdToCloneChild.His(n)
                    for elem in stammbaum:
                        Idx = self.mdbAccessor.GetIndex(elem)
                        Cf = self.mdbAccessor.GetContainingFeature(elem)
                        CfName = self.__Decode(Cf).name
                        cmdToCloneChild.Pth(CfName)
                        cmdToCloneChild.Idx(Idx)
                    cmdToCloneChild.Pth(cfName)
                    cmdToCloneChild.Idx(idx)
                    if cmd.cmd == 'SET':
                        cmpcmd.Set(cmdToCloneChild, cmd.a[1], cmd.a[2])
                    elif cmd.cmd == 'REM':
                        cmpcmd.Rem(cmdToCloneChild, cmd.a[1], cmd.a[2])
                    elif cmd.cmd == 'ADD':
                        cmpcmd.Add(cmdToCloneChild, cmd.a[1], cmd.a[2])
                # "normal" commands
                else:
                    res.append(cmd)
        self.cmdList.clear()
        # # make sure no changes will be made to "new"
        # for cmd in res:
        #     if cmd.cmd == 'CLO' or cmd.cmd == 'CMP':
        #         # print(self.TestSerializer.Ser(cmd))
        #         continue
        #     # print(self.mdbAccessor.GetAllParents(cmd.a[0]))
        #     if self.oldRoot not in self.mdbAccessor.GetAllParents(cmd.a[0]):
        #         name = 'noname'
        #         if hasattr(self.__Decode(cmd.a[0]), 'name') and self.__Decode(cmd.a[0]).name:
        #             name = self.__Decode(cmd.a[0]).name
        #         print("ERROR! Change made to new root:", cmd.a[0], name)
        #         print("Corresponding command:", self.logSerializer.Ser(cmd))
        #         raise EoqError('', '')
        return res
    
    # PRIVATE
    def __StartMergeRoutine(self):
        ''' 
        First function. Finds all differences and stores the corresponding
        commands in cmdList.
        If an object "A" has a reference "REF1" to an object "B" which has not been treated yet
        -> the treatment of "REF1" is postponed until after object "B" has been treted
        since object B could have been modified.
        '''
        start = timer()
        status = ResTypes.OKY
        # start recursive routine for objects
        if self.writeLog:
            self.logger.Info('Starting Diff.py for old: {} and new: {}'.format(self.oldRoot, self.newRoot))
            self.logger.Info('Level | Object name')
            self.logger.Info('         --- (all objects from old)')
            self.logger.Info('  -> is    | (old Object)     | Status  ')
        status = self.__MergeRoutine(self.oldRoot, self.newRoot)
        # set postponed references
        for ref in self.postponedRefs:
            old = ref[0]
            new = ref[1]
            # set the correct reference in new root
            [modified, newvalues] = self.__MergeReferences(old, new, donotpostpone = True)
            # later it will be updated to old
            addtolist = True
            for entry in self.refsToUpdate:
                if entry[0] == old:
                    addtolist = False
                    break
            if addtolist and modified and modified:
                self.refsToUpdate.append([old, modified, newvalues])
        self.postponedRefs.clear()
        # print("TIME after __MergeRoutine: {} s".format(timer()-start))
        return status

    def __StartUpdateReferences(self):
        ''' 
        Second function. Adds commands to cmdList to update 
        all the modified references from cloned / modified objects
        to be pointing to the old model.
        '''
        start = timer()
        status = ResTypes.OKY
        if self.writeLog:
            self.logger.Info('Updating references of new clones ...')
        # print(self.clonedObjs)
        for clone in self.clonedObjs:
            self.__UpdateCloneReferences(clone, clone)
        #
        if self.writeLog:
            self.logger.Info('Updating references of modified objects ...')
        # print(self.refsToUpdate)
        for entry in self.refsToUpdate:
            # print(entry)
            old = entry[0]
            modified = entry[1]
            newvalues = entry[2]
            # update reference command to old root
            self.__UpdateReferences(old, modified, newvalues)
        self.refsToUpdate.clear()
        # self.clonesToUpdate.clear()
        # print("TIME after __RefinishReferences: {} s".format(timer()-start))
        return status

    def __StartRemoveLoop(self):
        '''
        Third and last function. Finds all removed objects.
        '''
        start = timer()
        status = ResTypes.OKY
        # start recursive routine for objects
        if self.writeLog:
            self.logger.Info('Removing any deleted objects ... ')
        status = self.__RemoveLoop(self.oldRoot)
        # print("TIME after __RemoveLoop: {} s".format(timer()-start))
        return status

    # PRIVATE
    def __Decode(self, obj): 
        return self.mdbAccessor.vc.Dec(obj)

    def __MergeRoutine(self, old, new, n = 0):
        '''
        1 Compares two objects using a Similarity Index
        2 Finds and executes the corresponding action for these objects
        3 Restarts recursively for all children
        '''
        status = ResTypes.OKY
        allSI = []
        allObj = []
        allChanges = []
        allMoveflags = []
        # iterate over all first level children of "new"
        # get all children and their SI
        children = self.__GetAllChildrenInOneList(new)
        # __GetAllChildrenInOneList Function is necessary!!!
        # Check difference in "OAAM" --> "CON on CPIOM" for example
        for child in children:
            [SI, obj, changes] = self.mdbAccessor.CalculateSimilarityIndex(child, old) # x = 0 .. 1
            allSI.append(SI)
            allObj.append(obj)
            allChanges.append(changes)
        # check for moved objects (two objects with only I in changes)
        # for now, it detects only one move command
        for (child, SI, obj, changes) in zip(children, allSI, allObj, allChanges):
            moveflag = False
            for changesObj in changes:
                if len(changesObj) == 1 and changesObj[0] == 'I':
                    moveflag = True
                    break
            allMoveflags.append(moveflag)
        # find any move commands
        allMoveCmds = []
        for (child, SI, obj, changes, moveflag) in zip(children, allSI, allObj, allChanges, allMoveflags):
            moveCmd = []
            if moveflag and sum(allMoveflags) == 2:
                myMoveindex = changes.index(['I']) # first moved object
                myIndex = self.mdbAccessor.GetIndex(child) # my index in new
                otherIndex = self.mdbAccessor.GetIndex(obj[myMoveindex]) # my index in old
                if otherIndex == 1: # so we only do one move command in the end
                    moveCmd = [obj[myMoveindex], myIndex]
            allMoveCmds.append(moveCmd)
        # now decide what to do with each object
        for (child, SI, obj, changes, moveCmd) in zip(children, allSI, allObj, allChanges, allMoveCmds):            
            childType = self.__Decode(child).eClass.name
            childrenOfSameType = self.mdbAccessor.GetAllChildrenOfType(new, childType)
            if self.writeLog:
                info = '{:-5} | '.format(n)
                info += '{:8} | '.format(str(child))
                info += '{:16} of type '.format(self.__MergeGetObjectNameForOutput(child))
                info += '{:16}'.format(childType)
                self.logger.Info(info)
                info = ''
            # get obj with best SI 
            indexOfHighestSI = self.__GetMostSimilarObject(SI, changes)
            highestSI = SI[indexOfHighestSI]
            bestObj = obj[indexOfHighestSI] # object with highest SI
            mychanges = changes[indexOfHighestSI] # changes of obj with highest Si
            if moveCmd: # moved object
                if self.writeLog:
                    info += '  -> moved     | '
                    info += '{:8} | '.format(str(moveCmd[0]))
                    info += '{:32} | '.format(self.__MergeGetObjectNameForOutput(moveCmd[0]))
                self.cmdList.append(Cmd('MOV', [moveCmd[0], moveCmd[1]-1]))
                self.treatedObjs.append(moveCmd[0])
                if self.writeLog:
                    info += 'objects moved'
                    # self.__PrintAllSI(SI, obj, changes)
                    self.__PrintAllSI(SI, obj, changes)
                    self.logger.Info(info)
                continue # nothing else needs to be done with the moved object
            # main decision part - based on SI
            if highestSI > 0.99: 
                # --> the exact object exists (unmodified)
                if self.writeLog:
                    info += '  -> identical | '
                    info += '{:8} | '.format(str(bestObj))
                    info += '{:32} | '.format(self.__MergeGetObjectNameForOutput(bestObj))
                self.treatedObjs.append(bestObj)
                # note: children of children could have very well been modified
                # this is why __MergeRoutine must be called again for all children
                if self.__GetAllChildrenInOneList(child):
                    if self.writeLog:
                        info += 'checking subchildren ...'
                        # self.__PrintAllSI(SI, obj, changes)
                        self.logger.Info(info)
                    status = self.__MergeRoutine(bestObj, child, n = n + 1)
                else:
                    if self.writeLog:
                        info += 'nothing to be done'
                        # self.__PrintAllSI(SI, obj, changes)
                        self.logger.Info(info)
            elif (highestSI < 0.66
                or (highestSI > 0.66
                    and 'N' in changes[SI.index(highestSI)] 
                    and len(childrenOfSameType) > len(SI))): 
                # --> no comparable object exists (newly created)
                # Hammer-Methode: im Zweifel klonen und das andere löschen
                if self.writeLog:
                    info += '  -> new       | '
                    info += ' -       | '
                resClo = self.commandRunner.ExecClo([child, CloModes.DEP], self.tid)
                feature = str(self.__Decode(self.mdbAccessor.GetContainingFeature(child)).name)
                cmpcmd = Cmp()
                cmpcmd.Clo(child, CloModes.DEP)
                cmpcmd.Add(old, feature, His(-1))
                self.cmdList.append(cmpcmd)
                self.clonedObjs.append(child)
                addedObjs = self.mdbAccessor.GetAllContainmentValues(old)[self.mdbAccessor.GetAllContainmentNames(old).index(feature)]
                # self.cmdListClo.append(addedObjs[-1])
                del resClo
                if self.writeLog:
                    info += 'cloned'
                    self.__PrintAllSI(SI, obj, changes)
                    self.logger.Info(info)
            else: 
                # --> similar object exists (modified)
                if self.writeLog:
                    info += '  -> modified  | '
                    info += '{:8} | '.format(str(bestObj))
                    info += '{:32} | '.format(self.__MergeGetObjectNameForOutput(bestObj))
                self.treatedObjs.append(bestObj)
                if 'A' in mychanges:
                    if self.writeLog:
                        info += 'Merging Attributes '
                    self.__MergeAttributes(bestObj, child) # only merges first level
                if 'R' in mychanges:
                    if self.writeLog:
                        info += 'Merging References '
                    [modified, newvalues] = self.__MergeReferences(bestObj, child) # only merges first level
                    addtolist = True
                    for entry in self.refsToUpdate:
                        if entry[0] == bestObj:
                            addtolist = False
                            break
                    if addtolist and modified:
                        self.refsToUpdate.append([bestObj, modified, newvalues])
                if 'N' in mychanges:
                    if self.writeLog:
                        info += 'Merging Name '
                    self.__MergeName(bestObj, child)
                    # add bestObj to this list so that no reference that had its name changed
                    # will be changed back to the new one 
                    self.modifiedReferences.append(bestObj)
                if self.__GetAllChildrenInOneList(child):
                    if self.writeLog:
                        info += ', checking subchildren ...'
                        # print(', checking subchildren ...')
                        self.__PrintAllSI(SI, obj, changes)
                        self.logger.Info(info)
                    status = self.__MergeRoutine(bestObj, child, n = n + 1)
                else:
                    if self.writeLog:
                        info += ', nothing to be done'
                        self.__PrintAllSI(SI, obj, changes)
                        self.logger.Info(info)
        return status

    def __GetMostSimilarObject(self, SI, changes):
        '''
        Returns the index in SI of the most similar object 
        that has been returned by CalculateSI.
        '''
        # get highest SI from SI array (could be multiple ones!)
        bestSI = None # index of selected object
        bestChg = None
        m = max(SI)
        highestindex = [i for i, j in enumerate(SI) if j == m]
        # if it's just one: we're done
        if len(highestindex) == 1:
            return SI.index(m)
        # if there are multiple ones: we have to select one
        else:
            tmpSI = []
            tmpchg = []
            for index in highestindex:
                tmpSI.append(SI[index])
                tmpchg.append(changes[index])
            for (n, (obj, chg)) in enumerate(zip(tmpSI, tmpchg)):
                if not bestSI:
                    # has not been set yet
                    bestSI = obj
                    bestChg = chg
                    idx = highestindex[n]
                else:
                    # has been previoulsy set
                    if len(chg) == 1 and 'I' in chg:
                        # replace and quit (index modification supercedes all)
                        bestSI = obj
                        bestChg = chg
                        idx = highestindex[n]
                        if self.writeLog:
                            self.logger.Info('Selecting:{}, {}'.format(idx, bestSI))
                        break
                    # further action needed in case of two equally rated objs
            if not bestSI:
                # default
                bestSI = tmpSI[0]
            return idx

    def __MergeGetObjectNameForOutput(self,obj,default='<UNNAMED>'):
        '''
        Returns a name string for logging and output.
        '''
        Obj = self.__Decode(obj)
        if hasattr(Obj, 'name') and Obj.name:
            return Obj.name
        else:
            return default
 
    def __GetAllChildrenInOneList(self, obj):
        '''
        Returns all children of all reference types in one single list.
        '''
        allReferenceValues = self.mdbAccessor.GetAllReferenceValues(obj)
        allReferenceETypes = []
        for v in allReferenceValues:
            if isinstance(v, list):
                if len(v) > 0:
                    allReferenceETypes.append(self.__Decode(v[0]).eClass.name)
            else:
                if v is not None:
                    allReferenceETypes.append(self.__Decode(v).eClass.name)
        res = []
        for reference in allReferenceETypes:
            children = [e for e in self.mdbAccessor.GetAllChildrenInstanceOfClass(obj, reference) if self.mdbAccessor.GetParent(e) == obj]
            for child in children:
                res.append(child)
        # remove duplicates
        res2 = []
        for e in res:
            if e not in res2:
                res2.append(e)
        return res2

    def __MergeAttributes(self, old, new):
        '''
        Merges all differences in attributes from old to new.
        Does not consider children.
        '''
        allAttributes = self.mdbAccessor.GetAllAttributes(new)
        allAttributeNames = self.mdbAccessor.GetAllAttributeNames(new)
        allAttributeValues = self.mdbAccessor.GetAllAttributeValues(new)
        allOldAttributes = self.mdbAccessor.GetAllAttributes(old)
        allOldAttributeNames = self.mdbAccessor.GetAllAttributeNames(old)
        allOldAttributeValues = self.mdbAccessor.GetAllAttributeValues(old)
        # iterate new
        for (newAttr, newName, newValue) in zip(allAttributes, 
            allAttributeNames, allAttributeValues):
            # find matching attribute in old
            for (oldAttr, oldName, oldValue) in zip(allOldAttributes,
                allOldAttributeNames, allOldAttributeValues):
                found = False
                if oldName == newName:
                    found = True
                    if oldValue == newValue:
                        continue
                    else:
                        # print('Setting changed value of', oldName, 'from:', oldValue, '->', newValue)
                        # self.Set(old, oldName, newValue)
                        if self.writeLog:
                            info = '__MergeAttributes: '
                            info += 'Set: '
                            info += '{0}:"{1}" -> {2}'.format(old, oldName, newValue)
                            self.logger.Info(info)
                        self.cmdList.append(Set(old, oldName, newValue))


    def __MergeReferences(self, old, new, donotpostpone = False):
        '''
        Merges all differences in references from new to old.
        Does not consider children, 
        because after merging attributes/references the children are treated afterwards.
        Returns the indexes of the respective references (modified).
        '''
        # print('-- (updaterefs) __MergeReferences:', old, new)
        allReferences      = self.mdbAccessor.GetAllReferences(new)
        allReferenceNames  = self.mdbAccessor.GetAllReferenceNames(new)
        allReferenceValues = self.mdbAccessor.GetAllReferenceValues(new)
        allContainmentNames = self.mdbAccessor.GetAllContainmentNames(new)
        allOldReferences      = self.mdbAccessor.GetAllReferences(old)
        allOldReferenceNames  = self.mdbAccessor.GetAllReferenceNames(old)
        allOldReferenceValues = self.mdbAccessor.GetAllReferenceValues(old)
        modified = []
        newvalues = []
        # iterate new
        for (newRef, newName, newValue) in zip(allReferences,
            allReferenceNames, allReferenceValues):
            # do not change containments since they are 
            # treated as children in the next level
            if newName in allContainmentNames:
                continue
            # find matching Reference
            for (oldRef, oldName, oldValue) in zip(allOldReferences,
                allOldReferenceNames, allOldReferenceValues):
                if newName != oldName:
                    continue
                if oldValue in self.modifiedReferences:
                    # print('Continuing modifiedReferences...')
                    continue
                if self.__Decode(oldRef).upperBound == 1:
                    # set
                    newObj = self.__Decode(newValue)
                    oldObj = self.__Decode(oldValue)
                    if hasattr(newObj, 'name') and newObj.name:
                        newObjName = newObj.name
                    else:
                        newObjName = 'noname'
                    if hasattr(oldObj, 'name') and oldObj.name:
                        oldObjName = oldObj.name
                    else:
                        oldObjName = 'noname'
                    if newValue == None or oldValue == None:
                        continue
                    if not oldObjName == newObjName:
                        # skip untreated reference values
                        # must be done at the end instead
                        if (not donotpostpone) and (oldValue not in self.treatedObjs):
                            if self.writeLog:
                                info = 'Postponing __MergeReferences for: '
                                info += 'reference {0}:"{1}" value {2}:"{3}"'.format(newRef, 
                                    newName, newValue, newObjName)
                                self.logger.Info(info)
                            self.refsToUpdate.append([old, oldRef, newValue])
                            self.postponedRefs.append([old, new])
                            continue
                        # actual code
                        if self.writeLog:
                            info = '__MergeReferences: '
                            info += 'Set: '
                            info += '{0} {1}:"{2}" {3}:"{4}"'.format(old, 
                                oldRef, oldName, newValue, newObjName)
                            self.logger.Info(info)
                        self.cmdList.append(Set(old, oldName, newValue))
                        modified.append(oldRef)
                        newvalues.append(newValue)
                else:
                    # rem/add
                    for (newV, oldV) in zip(newValue, oldValue):
                        newObj = self.__Decode(newV)
                        oldObj = self.__Decode(oldV)
                        if hasattr(newObj, 'name') and newObj.name:
                            newObjName = newObj.name
                        else:
                            newObjName = 'noname'
                        if hasattr(oldObj, 'name') and oldObj.name:
                            oldObjName = oldObj.name
                        else:
                            oldObjName = 'noname'
                        if newV == None or oldV == None:
                            continue
                        if not oldObjName == newObjName:
                            # skip untreated reference values
                            # must be done at the end instead
                            if (not donotpostpone) and (oldV not in self.treatedObjs):
                                if self.writeLog:
                                    info = 'Postponing __MergeReferences for: '
                                    info += 'reference {0}:"{1}" value {2}:"{3}"'.format(newRef, newName, newV, newObjName)
                                    self.logger.Info(info)
                                self.refsToUpdate.append([old, oldRef, newV])
                                self.postponedRefs.append([old, new])
                                continue
                            if self.writeLog:
                                info = '__MergeReferences: '
                                info += 'Rem: '
                                info += '{0} {1}:"{2}" {3}:"{4}"'.format(old, 
                                    oldRef, oldName, oldV, oldObjName)
                                self.logger.Info(info)
                            self.cmdList.append(Rem(old, oldName, oldV))
                            if self.writeLog:
                                info = '__MergeReferences: '
                                info += 'Add: '
                                info += '{0} {1}:"{2}" {3}:"{4}"'.format(old, 
                                    oldRef, oldName, newV, newObjName)
                                self.logger.Info(info)
                            self.cmdList.append(Add(old, oldName, newV))
                            modified.append(oldRef)
                            newvalues.append(newV)
        if len(modified) == 1:
            modified = modified[0]
        if len(newvalues) == 1:
            newvalues = newvalues[0]
        return [modified, newvalues]

    def __MergeName(self, bestObj, child):
        childObj = self.__Decode(child)
        if hasattr(childObj, 'name') and childObj.name:
            name = childObj.name
            if self.writeLog:
                info = '__MergeName: '
                info += 'Set: '
                info += '{0} name -> {1}'.format(bestObj, name)
                self.logger.Info(info)
            self.cmdList.append(Set(bestObj, 'name', name))

    def __UpdateReferences(self, obj, modifiedObj, newvalues):
        '''
        Updates the references of modified references (which have their root in new) 
        such that they fit to old.
        '''
        # modifiedObj is either a single reference or a list
        if not isinstance(modifiedObj, list):
            modifiedObj = [modifiedObj]
        if not isinstance(newvalues, list):
            newvalues = [newvalues]
        assert len(modifiedObj) == len(newvalues)
        for (ref, newValue) in zip(modifiedObj, newvalues):
            name = 'noname'
            if hasattr(self.__Decode(obj), 'name') and self.__Decode(obj).name:
                name = self.__Decode(obj).name
            value = self.mdbAccessor.GetAllReferenceValues(obj)[self.mdbAccessor.GetAllReferences(obj).index(ref)]
            refname = self.mdbAccessor.GetAllReferenceNames(obj)[self.mdbAccessor.GetAllReferences(obj).index(ref)]
            # print('UpdateReferences for', obj, name, ':', ref, refname, value, ', newValue:', newValue)
            if not value:
                continue
            # set the reference
            if self.__Decode(ref).upperBound == 1:
                # set
                if self.newRoot not in self.mdbAccessor.GetAllParents(newValue):
                    raise EoqError('Warning: __FindCorrespondingObject got an object that is not rooted in new.', '')
                corrRef = self.__FindCorrespondingObject(newValue)
                # print(self.mdbAccessor.GetAllParents(value))
                if corrRef == newValue or not corrRef:
                    continue
                corrRefObj = self.__Decode(corrRef)
                corrRefName = 'noname'
                if hasattr(corrRefObj, 'name') and corrRefObj.name:
                    corrRefName = corrRefObj.name
                # print('Corresponding reference found:', corrRef, corrRefName)
                if self.writeLog:
                    info = '__UpdateReferences: '
                    info += 'Set: '
                    info += '{0} {1}:"{2}" {3}:"{4}"'.format(obj, 
                        ref, refname, corrRef, corrRefName)
                    self.logger.Info(info)
                self.cmdList.append(Set(obj, refname, corrRef))
            else:
                # rem/add
                # for v in newValue:
                # tbd multiple values
                if self.newRoot not in self.mdbAccessor.GetAllParents(newValue):
                    raise EoqError('Warning: __FindCorrespondingObject got an object that is not rooted in new.', '')
                corrRef = self.__FindCorrespondingObject(newValue)
                if corrRef == newValue or not corrRef:
                    continue
                corrRefObj = self.__Decode(corrRef)
                corrRefName = 'noname'
                if hasattr(corrRefObj, 'name') and corrRefObj.name:
                    corrRefName = corrRefObj.name
                if self.writeLog:
                    info = '__UpdateReferences: '
                    info += 'Rem: '
                    info += '{0} {1}:"{2}" {3}'.format(obj, 
                        ref, refname, newValue)
                    self.logger.Info(info)
                self.cmdList.append(Rem(obj, refname, newValue))
                if self.writeLog:
                    info = '__UpdateReferences: '
                    info += 'Add: '
                    info += '{0} {1}:"{2}" {3}:"{4}"'.format(obj, 
                        ref, refname, corrRef, corrRefName)
                    self.logger.Info(info)
                self.cmdList.append(Add(obj, refname, corrRef))

    def __UpdateCloneReferences(self, clone, originalClone):
        '''
        Adds all the references of new clones to the command list.
        Since we don't know the #ID of the clones, we create the commands as if they
        were applied to the original object. In GetCommands() this is corrected and a large
        Cmp-Command is created.
        '''
        # print('__UpdateCloneReferences for', clone, self.__Decode(clone).name)
        refs      = self.mdbAccessor.GetAllReferences(clone)
        refNames  = self.mdbAccessor.GetAllReferenceNames(clone)
        refValues = self.mdbAccessor.GetAllReferenceValues(clone)
        # print(refs, refNames, refValues)
        for (ref, name, value) in zip(refs, refNames, refValues):
            if not value:
                continue
            # print(ref, name, value)
            if self.__Decode(ref).upperBound == 1:
                # set
                corrObj = self.__FindCorrespondingObject(value)
                if (not corrObj) or (corrObj == value):
                    continue
                # print('__UpdateCloneReferences:', clone, 
                #     'corresponding:', corrObj, 
                #     self.__Decode(corrObj).name,
                #     self.__Decode(corrObj).eClass.name)
                self.cmdList.append(Set(clone, name, corrObj))
            else:
                # rem/add
                for v in value:
                    corrObj = self.__FindCorrespondingObject(v)
                    # print('__UpdateCloneReferences:', clone, 
                    #     'corresponding:', corrObj)
                    if (not corrObj) or (corrObj == v):
                        continue
                    self.cmdList.append(Rem(clone, name, v))
                    self.cmdList.append(Add(clone, name, corrObj))
        # needs to be done for all children of clone after cloning
        children = self.__GetAllChildrenInOneList(clone)
        if children:
            # print('Clone children need treatment:', clone, children)
            # save children so that the commands are changed to the actual clone
            # and not the object in new
            for child in children:
                self.cloneChildren.append(child)
                self.cloneList.append([child, originalClone])
                self.__UpdateCloneReferences(child, originalClone)

    def __FindCorrespondingObject(self, objToAnalyse):
        ''' 
        In: Any object in new.
        Out: The corresponding object in old.
        '''
        ObjToAnalyse = self.__Decode(objToAnalyse) # still in new
        objToAnalyseName = 'noname'
        if hasattr(ObjToAnalyse, 'name') and ObjToAnalyse.name:
            objToAnalyseName = ObjToAnalyse.name
        # print('- FindCorrespondingObject for', objToAnalyse, objToAnalyseName)
        valueName = ObjToAnalyse.eClass.name # name of the reference in question
        tree = self.mdbAccessor.GetAllParents(objToAnalyse)
        if tree[0] == self.oldRoot:
            return None
        correctTree = [self.oldRoot]
        for elem in tree:
            elemObj = self.__Decode(elem)
            elemEName = elemObj.eClass.name
            elemName = 'noname'
            if hasattr(elemObj, 'name') and elemObj.name:
                elemName = elemObj.name
            # continue if just workspace etc.
            if elemEName in ['Workspace', 'Directory', 'ModelResource']:
                continue
            # find elem in children of last entry of correctTree
            children = self.__GetAllChildrenInOneList(correctTree[-1])
            nextElem = None
            for child in children:
                obj = self.__Decode(child)
                objEName = obj.eClass.name
                name = 'noname'
                if hasattr(obj, 'name') and obj.name:
                    name = obj.name
                if elemEName == objEName and elemName == name:
                    nextElem = child
            if nextElem:
                correctTree.append(nextElem)
                continue
        # now we have the correct tree in old (completely in old!)
        # print('correctTree', correctTree) # tree is still fully in "new"
        # for t in correctTree:
        #     name = 'noname'
        #     tObj = self.__Decode(t)
        #     if hasattr(tObj, 'name') and tObj.name:
        #         name = tObj.name
        #     print(t, name, tObj.eClass.name)
        # find now the correct reference
        correctRefChildren = self.__GetAllChildrenInOneList(correctTree[-1])
        # print('correct ref children:', correctRefChildren)
        # for crc in correctRefChildren:
        #     crcObj = self.__Decode(crc)
        #     name = 'noname'
        #     if hasattr(crcObj, 'name') and crcObj.name:
        #         name = crcObj.name
        #     print('-', crc, name, crcObj.eClass.name)
        # get all references that have the same name and type
        correctReferences = []
        for child in correctRefChildren:
            childObj = self.__Decode(child)
            childName = 'noname'
            if hasattr(childObj, 'name') and childObj.name:
                childName = childObj.name
            childECLass = childObj.eClass.name
            # print('-', child, childName, childECLass)
            if childName == ObjToAnalyse.name and childECLass == ObjToAnalyse.eClass.name:
                correctReferences.append(child) 
        # decide on which reference to take
        correctReferenceFinal = None
        # print('- ', correctReferences)
        for correctReference in correctReferences:
            # print('- correctReferences:', correctReference,
            #     self.__Decode(correctReference).name,
            #     self.__Decode(correctReference).eClass.name)
            # TBD!
            correctReferenceFinal = correctReferences[0]
        return correctReferenceFinal

    def __RemoveLoop(self, old):
        '''
        Removes all objects from 'Old' that have not been found in 'New'.
        '''
        #print(self.clonesToUpdate)
        #print(self.clonedObjs)
        status = ResTypes.OKY
        childrenOld = self.__GetAllChildrenInOneList(old)
        if not childrenOld:
            return status
        for obj in childrenOld:
            if obj not in self.treatedObjs and obj not in self.clonedObjs:
                objOld = self.__Decode(obj)
                eFeature = objOld.eContainmentFeature()
                if eFeature is None: # to be deleted as soon as problem with duplicates is resolved
                    # print('empty eFeature', obj)
                    continue
                if self.writeLog:
                    info = 'removing object: ' + str(obj) + ' ' 
                    info += self.__MergeGetObjectNameForOutput(objOld) 
                    info += ', of type: ' + eFeature.name
                    parent = self.mdbAccessor.GetParent(obj)
                    info += ', with parent: ' + self.__MergeGetObjectNameForOutput(parent) 
                    info += ' ' + str(parent)
                    self.logger.Info(info)
                self.cmdList.append(Rem(old, eFeature.name, obj))
            elif obj in self.clonedObjs:
                # do not go deeper, because cloned subobjects do not appear in clonedObjs
                continue
            else:
                self.__RemoveLoop(obj)
        return status

    def __PrintAllSI(self, SIArray, objArray, allChangesArray):
        # settings
        for (SI, obj, changes) in zip(SIArray, objArray, allChangesArray):
            info = '         --- {:16s} - {:8s}:'.format(self.__MergeGetObjectNameForOutput(self.__Decode(obj)), str(obj))
            info += 'SI = {:4.2f} - modifications: '.format(SI)
            for (i, change) in enumerate(changes):
                if i < (len(changes) - 1):
                    info += '{:1},'.format(change)
                else:
                    info += '{:1}'.format(change)
            if not changes:
                info += 'none'
            self.logger.Info(info)

    def __SubroutineForCreateObjectOverviewFile(self, tempList, root):
        ''' Recursive subroutine for CreateObjectOverviewFile. '''
        children = self.__GetAllChildrenInOneList(root)
        for child in children:
            name = 'noname'
            obj = self.__Decode(child)
            if hasattr(obj, 'name') and obj.name:
                name = obj.name
            if [child, name] not in tempList:
                tempList.append([child, name])
            if self.__GetAllChildrenInOneList(child):
                self.__SubroutineForCreateObjectOverviewFile(tempList, child)





'''
 Legacy support for old jseoq commands
 2019 Bjoern Annighoefer
'''

from ..util import EoqError
from ..command.command import CmdTypes,Cmd,Cal,Asc,CloModes
from ..command.result import ResTypes
from ..query.query import QrySegTypes,Qry,Seg


#import legacy stuff
from eoq1.model import CommandTypesE, ResultTypesE, ValueTypesE, SegmentTypesE, IndexTypesE, OperatorTypesE, ErrorResult, CompoundResult, HelloResult, GoodbyeResult, SessionResult, StatusResult, ChangesResult, RetrieveResult, UpdateResult, CreateResult, CloneResult, CallResult, AsyncCallResult, AbortCallResult, CallStatusResult, CloneModesE


def UpgradeCmd(legacyCmd):
    cloModeConversion = {}
    cloModeConversion[CloneModesE.CLASS] = CloModes.CLS
    cloModeConversion[CloneModesE.ATTRIBUTES] = CloModes.ATT
    cloModeConversion[CloneModesE.DEEP] = CloModes.DEP
    cloModeConversion[CloneModesE.FULL] = CloModes.FUL
    
 
    
    cmd = None
    if(legacyCmd.type == CommandTypesE.HELLO):
        cmd = Cmd(CmdTypes.HEL,[UpgradeVal(legacyCmd.user),UpgradeVal(legacyCmd.identification)])
    elif(legacyCmd.type == CommandTypesE.GOODBYE):
        cmd = Cmd(CmdTypes.GBY,[UpgradeVal(legacyCmd.sessionId)])
    elif(legacyCmd.type == CommandTypesE.SESSION):
        cmd = Cmd(CmdTypes.SES,[UpgradeVal(legacyCmd.sessionId)])
    elif(legacyCmd.type == CommandTypesE.STATUS):
        cmd = Cmd(CmdTypes.STS,[])
    elif(legacyCmd.type == CommandTypesE.CHANGES):
        cmd = Cmd(CmdTypes.CHG,[UpgradeVal(legacyCmd.earliestChangeId),0])
    elif(legacyCmd.type == CommandTypesE.RETRIEVE):
        cmd = Cmd(CmdTypes.GET,UpgradeQry(legacyCmd.target,legacyCmd.query))
    elif(legacyCmd.type == CommandTypesE.CREATE):
        cmd = Cmd(CmdTypes.CRN,[UpgradeVal(legacyCmd.packageNsUri),UpgradeVal(legacyCmd.className),UpgradeVal(legacyCmd.n)])
    elif(legacyCmd.type == CommandTypesE.UPDATE):
        cmd = Cmd(CmdTypes.SET,[UpgradeQry(legacyCmd.target,legacyCmd.query),UpgradeVal(legacyCmd.value)])
        #cmd = Cmd(CmdTypes.ADD,[legacyCmd.target,legacyCmd.value])
        #cmd = Cmd(CmdTypes.REM,[legacyCmd.target,legacyCmd.value])
    elif(legacyCmd.type == CommandTypesE.CLONE):
        cmd = Cmd(CmdTypes.CLO,[UpgradeVal(legacyCmd.target),cloModeConversion[legacyCmd.mode]])
    elif(legacyCmd.type == CommandTypesE.CALL):
        cmd = Cal(UpgradeVal(legacyCmd.action),UpgradeVal(legacyCmd.args))
    elif(legacyCmd.type == CommandTypesE.ASYNCCALL):
        cmd = Asc(UpgradeVal(legacyCmd.action),UpgradeVal(legacyCmd.args))
    elif(legacyCmd.type == CommandTypesE.CALLSTATUS):
        cmd = Cmd(CmdTypes.GBY,[UpgradeVal(legacyCmd.callId)])
    elif(legacyCmd.type == CommandTypesE.ABORTCALL):
        cmd = Cmd(CmdTypes.GBY,[UpgradeVal(legacyCmd.callId)])
    elif(legacyCmd.type == CommandTypesE.COMPOUND):
        subcommands = []
        for lsc in legacyCmd.commands:
            sc = UpgradeCmd(lsc) 
            subcommands.append(sc)
        
        cmd = Cmd(CmdTypes.CMP,subcommands)
    else:
        raise EoqError(0,'Unknown command type: '+legacyCmd.type)
    return cmd


def UpgradeQry(target,legacyQry):
    root = UpgradeVal(target)
    qry = Qry(root)
    for ls in legacyQry.segments:
        if(ls.type == SegmentTypesE.PATH):
            name = ls.identifier
            qry.Pth(name)
        elif(ls.type == SegmentTypesE.CLAZZ):
            name = ls.identifier
            qry.Cls(name)
        elif(ls.type == SegmentTypesE.INSTANCE):
            name = ls.identifier
            qry.Ino(name)
        elif(ls.type == SegmentTypesE.META):
            name = ls.identifier
            qry.Met(name)
        else:
            raise EoqError(0,"Unknown segment type: "+ls.type)
        

        #consider selectors 
        if(ls.selector):
            selector = ls.selector
            subqry = Qry()
            subqry.Pth(selector.name)
            if(selector.operator.type == OperatorTypesE.EQUAL):
                    subqry.Equ(UpgradeVal(selector.value))
            elif(selector.operator.type == OperatorTypesE.NOTEQUAL):
                    subqry.Neq(UpgradeVal(selector.value))
            elif(selector.operator.type == OperatorTypesE.GREATER):
                    subqry.Gre(UpgradeVal(selector.value))
            elif(selector.operator.type == OperatorTypesE.LESS):
                    subqry.Les(UpgradeVal(selector.value))
            else:
                raise EoqError(0,"Unknown operator in selector: %s"%(selector.operator))
            
            qry.Sel(subqry)
        

        #consider index
        if(ls.index):
            index = ls.index
            if(index.type == IndexTypesE.NUMBER):
                qry.Idx(index.value)
            else:
                raise EoqError(0,"Legacy conversion does not support index type: %s"%(index.type))
    return qry


def UpgradeVal(legacyVal):
    qry = None
    if(legacyVal.type == ValueTypesE.LIST):
        subqrys = []
        for sv in legacyVal.v:
            sq = UpgradeVal(sv)
            subqrys.append(sq)
        
        #qry = Seg(QrySegTypes.ARR,subqrys)
        qry = subqrys #is that translation correct?
    elif(legacyVal.type == ValueTypesE.INT):
        qry = legacyVal.v
    elif(legacyVal.type == ValueTypesE.FLOAT):
        qry = legacyVal.v
    elif(legacyVal.type == ValueTypesE.BOOL):
        qry = legacyVal.v
    elif(legacyVal.type == ValueTypesE.STRING):
        qry = legacyVal.v
    elif(legacyVal.type == ValueTypesE.OBJECTREF):
        qry = Seg(QrySegTypes.OBJ,legacyVal.v)
    elif(legacyVal.type == ValueTypesE.EMPTY):
        qry = None
    elif(legacyVal.type == ValueTypesE.HISTORYREF):
        qry = Seg(QrySegTypes.HIS,legacyVal.v)
    else:
        raise EoqError(0,"Unknown value type: %s"%(legacyVal.type))
    return qry


def DowngradeRes(res):
    legacyRes = None
    if(res.res == CmdTypes.CMP):
        legacyRes = CompoundResult()
        if(res.s == ResTypes.ERR):
            legacyRes.type = ResultTypesE.COMPOUND_ERROR #if a single command failed the full compound command is failed.
        else:
            for subresult in res.v:
                lsr = DowngradeRes(subresult)
                legacyRes.results.append(lsr)
        
    else: # single commands
        if(res.res == CmdTypes.HEL):
            legacyRes = HelloResult()
            legacyRes.sessionId = res.v
        elif(res.res == CmdTypes.GBY):
            legacyRes = GoodbyeResult()
        elif(res.res == CmdTypes.SES):
            legacyRes = SessionResult()
        elif(res.res == CmdTypes.STS):
            legacyRes = StatusResult()
            legacyRes.changeId = res.v
        elif(res.res == CmdTypes.CHG):
            legacyRes = ChangesResult()
            legacyRes.changes = DowngradeVal(res.v) #could be more complicated if changes format differs in future
        elif(res.res == CmdTypes.GET):
            legacyRes = RetrieveResult()
            legacyRes.value = DowngradeVal(res.v)
        elif(res.res == CmdTypes.CRN):
            legacyRes = CreateResult()
            legacyRes.value = DowngradeVal(res.v)
        elif(res.res == CmdTypes.SET):
            legacyRes = UpdateResult()
            legacyRes.target = DowngradeVal(res.v[0])
        elif(res.res == CmdTypes.ADD):
            legacyRes = UpdateResult()
            legacyRes.target = DowngradeVal(res.v[0])
        elif(res.res == CmdTypes.REM):
            legacyRes = UpdateResult()
            legacyRes.target = DowngradeVal(res.v[0])
        elif(res.res == CmdTypes.MOV):
            legacyRes = UpdateResult()
            legacyRes.target = DowngradeVal(res.v[0])
        elif(res.res == CmdTypes.CLO):
            legacyRes = CloneResult()
            legacyRes.value = DowngradeVal(res.v)
        elif(res.res == CmdTypes.CAL):
            legacyRes = CallResult()
            legacyRes.callId = 0 #res.v[0]
            legacyRes.returnValues = DowngradeVal(res.v) #skip the first entry
        elif(res.res == CmdTypes.ASC):
            legacyRes = AsyncCallResult()
            legacyRes.callId = res.v
        elif(res.res == CmdTypes.CST):
            legacyRes = CallStatusResult()
            legacyRes.callId = res.v[0]
            legacyRes.callStatus = res.v[1]
            legacyRes.result = DowngradeVal(res.v[1:]) #skip the first entry
        elif(res.res == CmdTypes.ABC):
            legacyRes = AbortCallResult()
        else:
            raise EoqError(0,"Unknown result type: %s"%(res.res))
        
        
        #check if an error happened then we must override the legacy result
        if(res.s == ResTypes.ERR):
            commandType = legacyRes.commandType
            legacyRes = ErrorResult()
            legacyRes.commandType = commandType
            legacyRes.code = 0
            legacyRes.message = res.v
         

        # copy the transaction id
        legacyRes.transactionId = res.n #works for all results
    return legacyRes


from eoq1.model import IntValue, FloatValue, BoolValue, StringValue, ObjectRefValue, HistoryRefValue, EmptyValue, ListValue

def DowngradeVal(qry):
    legacyVal = 0
    if(isinstance(qry,bool)): #bool must before int, because bool is also in int in Python
        legacyVal = BoolValue()
        legacyVal.v = qry
    elif(isinstance(qry,int)):
        legacyVal = IntValue()
        legacyVal.v=qry
    elif(isinstance(qry,float)):
        legacyVal = FloatValue()
        legacyVal.v = qry
    elif(isinstance(qry,str)):
        legacyVal = StringValue()
        legacyVal.v = qry
    elif(isinstance(qry,list)):
        legacyVal = ListValue()
        for sv in qry:
            legacyVal.v.append(DowngradeVal(sv))
        return legacyVal
    elif(None == qry):
        legacyVal = EmptyValue()
    elif(qry.qry == QrySegTypes.OBJ):
        legacyVal = ObjectRefValue()
        legacyVal.v = qry.v
    elif(qry.qry == QrySegTypes.HIS):
        value = HistoryRefValue()
        value.v = qry.v
    else:
        raise EoqError(0,'Cannot convert data type %s to legacy EOQ value: %s'%(type(qry).__name__,qry))
    return legacyVal

from ..domain import Domain
from ..serialization import TextSerializer
from ..util.logger import NoLogging,LogLevels

class LegacyDomain(Domain):
    def __init__(self,domain,serializer=TextSerializer(),logger=NoLogging()):
        self.domain = domain
        self.serializer = serializer
        self.logger = logger
        
    def RawDo(self,legCmd,sessionId=None): #sessionId is not used
        cmd = UpgradeCmd(legCmd)
        self.logger.PassivatableLog(LogLevels.INFO,lambda : "legacy conversion to: %s"%(self.serializer.serialize(cmd)))
        res = self.domain.RawDo(cmd)
        legRes = DowngradeRes(res)
        return legRes
    
from ..frame import FrameHandler,FrameTypes
from eoq1 import CommandParser,ResultParser

import traceback
    
class LegacyTextFrameHandler(FrameHandler):
    def __init__(self,domain,logger=NoLogging()):
        super().__init__()
        self.domain = domain
        self.logger = logger
        
    def Handle(self,frames,sessionId=None):
        for i in range(len(frames)):
            try:
                if(frames[i].eoq == FrameTypes.CMD):
                    cmdStr = frames[i].dat
                    cmd = CommandParser.StringToCommand(cmdStr)
                    res = self.domain.RawDo(cmd,sessionId)
                    resStr = ResultParser.ResultToString(res)
                    frames[i].dat = resStr
                    frames[i].eoq = FrameTypes.RES
                else: 
                    raise EoqError(0,"Can not handle frame type: %s"%(frames[i].eoq))
            except Exception as e:
                errorMsg = "Got invalid frame: %s"%(str(e))
                self.logger.Error(errorMsg)
                traceback.print_exc();
                #print("WARNING: %s"%(errorMsg))
                frames[i].eoq = FrameTypes.ERR
                frames[i].dat = errorMsg
        return frames





'''
 2019 Bjoern Annighoefer
'''

class EvtTypes:
    CHG = "CHG" # change
    WAT = "WAT" # watch event (not implemented)
    OUP = "OUP" # output of a call
    INP = "INP" # input for a call
    CST = "CST" # call status change
    CVA = "CVA" # call value change event
    MSG = "MSG" # message
    CUS = "CUS" # custom event with user defined structure
    VCA = "VCA" # validation constraint added
    VCR = "VCR" # validation constraint removed
    VCS = "VCS" # validation constraint status
    
class ChgTypes:
    SET = 'SET'
    ADD = 'ADD'
    REM = 'REM'
    MOV = 'MOV'
    ADC = 'ADC'
    RMC = 'RMC'
    
    
ALL_EVENT_TYPES = [EvtTypes.CHG,EvtTypes.OUP,EvtTypes.INP,EvtTypes.CST,EvtTypes.CVA,EvtTypes.MSG,EvtTypes.VCA,EvtTypes.VCR,EvtTypes.VCS]
    
        
class Evt:
    def __init__(self,ctype,key,args):
        self.evt = ctype
        self.k = key
        self.a = args  
  
class ChgEvt(Evt):
    def __init__(self,cid,ctype,target,feature,newVal=None,oldVal=None,oldOwner=None,oldFeature=None,oldIndex=None,tid=0,user='',sessionNumber=0):
        key = '%s:%s:%s'%(target,ctype,feature)
        super().__init__(EvtTypes.CHG,key,[cid,ctype,target,feature,newVal,oldVal,oldOwner,oldFeature,oldIndex,tid,user,sessionNumber])
        
class OupEvt(Evt):
    def __init__(self,callId,channelName,data):
        key = str(callId)
        super().__init__(EvtTypes.OUP,key,[callId,channelName,data])
        
class InpEvt(Evt):
    def __init__(self,callId,channelName,data):
        key = str(callId)
        super().__init__(EvtTypes.INP,key,[callId,channelName,data])
        
class CstEvt(Evt):
    def __init__(self,callId,status,info=''):
        key = str(callId)
        super().__init__(EvtTypes.CST,key,[callId,status,info])
        
class CvaEvt(Evt):
    def __init__(self,callId,value):
        key = str(callId)
        super().__init__(EvtTypes.CVA,key,[callId,value])
        
class MsgEvt(Evt):
    def __init__(self,key,msg):
        super().__init__(EvtTypes.MSG,key,msg)
        
class CusEvt(Evt):
    def __init__(self,key,data):
        super().__init__(EvtTypes.CUS,key,data)

class VcaEvt(Evt):
    def __init__(self,cid,ctype,target,feature,constraint):
        key = '%s:%s:%s'%(target,ctype,feature)
        super().__init__(EvtTypes.VCA,key,[cid,ctype,target,feature,constraint])

class VcrEvt(Evt):
    def __init__(self,cid,ctype,target,feature,constraint):
        key = '%s:%s:%s'%(target,ctype,feature)
        super().__init__(EvtTypes.VCA,key,[cid,ctype,target,feature,constraint])

class VcsEvt(Evt):
    def __init__(self,cid,ctype,target,feature,constraint):
        key = '%s:%s:%s'%(target,ctype,feature)
        super().__init__(EvtTypes.VCA,key,[cid,ctype,target,feature,constraint])
        
        
import traceback
import threading

class ObserverInfo:
    def __init__(self,callback,eventTypes,context,sessionId):
        self.callback = callback #the callback function of the event listener
        self.eventTypes = eventTypes #the events that are desired
        self.context = context #a unique identifier or None
        self.sessionId = sessionId #set if this observer is linked to a session
        
'''
    EventProvider
'''
class EvtProvider:
    def __init__(self):
        self.observers = {} # callback -> event
        self.eventQueu = [] #events reserved for the next notification
        #self.eventQueueMutex = threading.Lock()
        self.observerQueuMutex = threading.Lock()
        
    def Observe(self,callback,eventTypes=ALL_EVENT_TYPES,context=None,sessionId=None): #by default register for all events
        observerInfo = ObserverInfo(callback,eventTypes,context,sessionId)
        self.observerQueuMutex.acquire()
        try:
            self.observers[(callback,context)] = observerInfo
        finally:
            self.observerQueuMutex.release()
        
    def Unobserve(self,callback,context=None):
        self.observerQueuMutex.acquire()
        try:
            self.observers.pop((callback,context))
        finally:
            self.observerQueuMutex.release()
            
#     '''
#         EVENT FORWARDING
#     '''
#         
#     def ObserveExternalProvider(self,eventProvider,eventTypes=ALL_EVENT_TYPES,context=None):
#         eventProvider.Observe(self._OnExternalEvent,eventTypes,context)
#         
#     def UnobserveExternalProvider(self,eventProvider,context=None):
#         eventProvider.Unobserve(self._OnExternalEvent,context)
#     
#     def _OnExternalEvent(self,evts,src):
#         self.NotifyObservers(evts,src)
        
#     def QueuEvent(self,evt,src):
#         self.eventQueu.append(evt)

    '''
        INTERNAL EVENT PROCESSING
    '''

    def IsEventDesired(self,evt,callback,eventTypes,context,sessionId):
        return evt.evt in eventTypes
        
    def NotifyObservers(self,evts,excludedCallback=None,excludedContext=None): #sends multiple events 
        
        #copy the events from the internal list and release it
#         self.eventQueueMutex.acquire()
#         try:
#             newEvts = self.eventQueu + evts
#             self.eventQueu.clear()
#         finally:
#             self.eventQueueMutex.release()

        newEvts = evts.copy() #make a copy to ensure that the list is not changed outside when the event notification loop runs.
            
        
        self.observerQueuMutex.acquire()
        try:
            for observer in self.observers.items():
                observerInfo = observer[1]
                if(excludedCallback==observerInfo.callback and excludedContext==observerInfo.context):
                    continue
                filterdEvts = [e for e in newEvts if self.IsEventDesired(e,observerInfo.callback,observerInfo.eventTypes,observerInfo.context,observerInfo.sessionId)]
                #if(evt.evt in item[1] and excludedObserver!=item[0]):
                try:
                    if(0<len(filterdEvts)):
                        observerInfo.callback(filterdEvts,self)
                except:
                    print("EvtProvider: Warning observer callback failed:")
                    traceback.print_exc()
        finally:
            self.observerQueuMutex.release()
        


'''
2019 Bjoern Annighoefer
'''
from .error import EoqError

DATE_TIME_STR_FORMAT = "%m/%d/%Y %H:%M:%S"

'''
    A class indicating that nothing more to process here
'''
class Terminator:
    def __init__(self,v):
        self.v = v
    def __eq__(self, other):
        if(isinstance(other, Terminator)):
            return self.v == other.v
        else:
            return self.v == other

def Determinate(res):
    if(IsList(res)):
        res = [Determinate(r) for r in res]
    elif(isinstance(res, Terminator)):
        res = res.v
    return res

'''
    Operations on multiple objects, structures and lists
    
'''

def ApplyToAllElements(context,functor):
    res = None
    if(IsList(context)):
        res = [ApplyToAllElements(c,functor) for c in context]
    elif(isinstance(context, Terminator)):
        res = res
    else:
        res = functor(context)
    return res

#Applies and operation specified by the functor to all elements of arbitrary
#nested lists as long as they exhibit the same structure
# def ApplyToAllCorrespondingElements(a,b,functor):
#     res = None
#     if(IsList(a)):
#         res = [ApplyToAllCorrespondingElements(a[i],b[i],functor) for i in range(len(a))]
#     else:
#         res = functor(a,b) 
#     return res

def ApplyToAllElementsInA(a,b,functor):
    res = None
    if(IsList(a)):
        res = [ApplyToAllElementsInA(c,b,functor) for c in a]
    elif(isinstance(a, Terminator)):
        res = a
    else:
        res = functor(a,b)
    return res

def ApplyToAllElementsInB(a,b,functor):
    res = None
    if(IsList(b)):
        res = [ApplyToAllElementsInB(a,c,functor) for c in b]
    else:
        res = functor(a,b)
    return res

def ApplyToAllListsOfElementsInA(a,b,functor):
    res = None
    if(IsListOfObjects(a)):
        res = functor(a,b)
    else:
        res = [ApplyToAllListsOfElementsInA(c,b,functor) for c in a]
    return res

def ApplyToAllListsOfElementsInB(a,b,functor):
    res = None
    if(IsListOfObjects(b)):
        res = functor(a,b)
    else:
        res = [ApplyToAllListsOfElementsInB(a,c,functor) for c in b]
    return res

def ApplyToSimilarListsOfObjects(op1,op2,listVsListFunc,listVsStructFunc,structVsListOp,param=None):
    res = None
    if(IsListOfObjects(op1) and IsListOfObjects(op2)):
        res = listVsListFunc(op1,op2,param)
    elif(IsListOfObjects(op1)):
        res = listVsStructFunc(op1,op2,param)
    elif(IsListOfObjects(op2)):
        res = structVsListOp(op1,op2,param)
    elif(len(op1) == len(op2)):
        res = [ApplyToSimilarListsOfObjects(op1[i],op2[i],listVsListFunc,listVsStructFunc,structVsListOp,param) for i in range(len(op1))]
    else:
        raise EoqError(0,"Non comparable element list structures detected.")
    return res

def ApplyToSimilarElementStrutures(op1,op2,elemVsElemFunc,elemVsStruct,structVsElemOp,param=None):
    res = None
    if(IsNoList(op1) and IsNoList(op2)):
        res = elemVsElemFunc(op1,op2,param)
    elif(IsNoList(op1)):
        res = elemVsStruct(op1,op2,param)
    elif(IsNoList(op2)):
        res = structVsElemOp(op1,op2,param)
    elif(len(op1) == len(op2)):
        res = [ApplyToSimilarElementStrutures(op1[i],op2[i],elemVsElemFunc,elemVsStruct,structVsElemOp,param) for i in range(len(op1))]
    else:
        raise EoqError(0,"Non comparable element structures detected.")
    return res

def IsList(obj):
    return (isinstance(obj,list) and not isinstance(obj,str))

def IsNoList(obj):
    return (not isinstance(obj,list) or isinstance(obj,str))

def IsListOfObjects(obj):
    if(IsList(obj)):
        if(len(obj)==0): 
            return True
        for o in obj:
            if IsNoList(o):
                return True #any non list object make the search fail
    return False
    #return (IsList(obj) and (len(obj)==0 or IsNoList(obj[0])))

def ShowProgress(progress):
    print('Total progress: %d%%'%(progress))
    return



'''
 2019 Bjoern Annighoefer
'''

from ..event import EvtProvider

'''
  CALL MANAGER
'''

class CallManager(EvtProvider):
    def __init__(self,callId=-1):
        super().__init__()
        self.callId = callId
        
    def RegisterAction(self,name,action,handler):
        pass
        
    def UnregisterAction(self,name):
        pass
        
    def GetAllActions(self):
        pass
       
    def ChangeCallStatus(self,callId,status,info=''):
        pass
            
    def SetCallValue(self,callId,value):
        pass
            
    def AddCallOutputRaw(self,callId,channelName,data):
        pass
            
    def AddCallOutput(self,channelName,data):
        self.AddCallOutputRaw(self.callId,channelName,data)
            
    def RunCallSync(self,name,args,opts,tid,sessionId=None):
        pass

    def RunCallAsnyc(self,name,args,opts,tid,sessionId=None,parentSessionId=None):
        pass
    
    def AbortCall(self,callId,tid):
        pass
        



'''
Provides various eoq-related model functions
2020 Institute of Aircraft Systems, University of Stuttgart
'''

from eoq2 import Get,Set,Add,Rem,Clo,Crn,CloModes
from eoq2 import Qry,Pth
from eoq2.util import error
from eoq2.event import EvtTypes


def GetName(domain,item) -> str:
    """ Retrieves name for given item """
    return domain.Do(Get(Qry(item).Pth("name")))


def GetId(domain,item) -> str:
    """ Retrieves ID for given item """
    return domain.Do(Get(Qry(item).Pth("id")))


def EventLogger(evts,source):
    """ Creates EventLogger and handles logging """
    for evt in evts:
        if evt.evt == EvtTypes.CST:
            print("EVT: Status of call %d changed to %s" % (evt.a[0],evt.a[1]))
        elif evt.evt == EvtTypes.OUP:
            print("EVT: Output of call %d on channel %s: %s" % (evt.a[0],evt.a[1],evt.a[2]))
        elif evt.evt == EvtTypes.CVA:
            print("EVT: Result of call %d: %s" % (evt.a[0],evt.a[1]))
        elif evt.evt == EvtTypes.CHG:
            print("EVT: Change (%d): %s t: %s, f: %s, n: %s, [was: v:%s, o: %s, f:%s, i:%s] (tid:%d)" % (
                evt.a[0],evt.a[1],evt.a[2],evt.a[3],evt.a[4],evt.a[5],evt.a[6],evt.a[7],evt.a[8],evt.a[9]))
        elif evt.evt == EvtTypes.MSG:
            print("EVT: Message: %s" % evt.a)


def CreateNewModelResource(domain,outFile,outDir,modelRoot):
    """ Saves OAAM model resource to file """
    print('Saving OAAM model to %s...' % outFile,end='')
    try:
        outModel = domain.Do(Get(Qry(outDir).Pth('resources').Sel(Pth('name').Equ(outFile)).Idx(0)))
        print('Replaced existing %s' % outFile)
    except (IndexError,error.EoqError):
        outModel = domain.Do(Crn('http://www.eoq.de/workspacemdbmodel/v1.0','ModelResource',1))
        domain.Do(Set(Qry(outModel),'name',outFile))
        domain.Do(Add(Qry(outDir),'resources',outModel))
        print('Created new %s' % outFile)
    domain.Do(Rem(Qry(outModel),'contents',Qry(outModel).Pth('contents')))  # clear the contents
    domain.Do(Add(Qry(outModel),'contents',modelRoot))
    print('Saved.')
    return


def CloneModelResource(domain,tempFile,templateDir=None):
    """ Clones model resource from template """
    if templateDir:
        templateRoot = domain.Do(Get(Pth('subdirectories').Sel(Pth('name').Equ(templateDir)).Idx(0).Pth('resources')
                                     .Sel(Pth('name').Equ(tempFile)).Idx(0).Pth('contents').Idx(0)))  # workspace mdb
    else:
        templateRoot = domain.Do(Get(Pth('resources').Sel(Pth('name').Equ(tempFile)).Idx(0).Pth('contents').Idx(0)))
    return domain.Do(Clo(Qry(templateRoot),CloModes.FUL))



'''
 2019 Bjoern Annighoefer
'''

from ..cmdrunnerbaseddomain import CmdRunnerBasedDomain
from ...command.commandrunner import CmdRunner
from ...util.logger import NoLogging
from ...serialization import TextSerializer

class LocalMdbDomain(CmdRunnerBasedDomain):
    def __init__(self,mdbAccessor,maxChanges=100,logger=NoLogging(),serializer=TextSerializer(),enableBenchmark=False):
        self.mdbAccessor = mdbAccessor
        self.cmdRunner = CmdRunner(mdbAccessor,maxChanges=maxChanges,logger=logger,enableBenchmark=enableBenchmark)
        super().__init__(self.cmdRunner,logger=logger,serializer=serializer,enableBenchmark=enableBenchmark)
        
        
        
        


from .. import ValueParser,QueryParser,CommandParser
from ..model import *
from ..error import EoqError,EoqRetrieveError,EoqCreateError,EoqUpdateError,EoqContextError,EoqTransactionError,EoqIdError,EoqGeneralError,EoqActionError

from pyecore.resources import Resource,ResourceSet,URI
from pyecore.ecore import EOrderedSet,EObject,EPackage,EEnumLiteral,EClass,EAttribute,EReference,EProxy,MetaEClass
import pyecore.behavior as behavior # We need to import the 'behavior' package

import types
import os
import time
import glob
import sys,traceback
from datetime import datetime
import shutil
import logging


import threading #make local domain threadsafe
from pyeoq.error.error import EoqValueError
from pip._vendor.colorama.ansi import Back
from pyeoq.resultparser import ResultParser

#Implement methods
@PathElementA.behavior
def actualPath(self):
	path = self.name
	parent = self.eContainer()
	while(isinstance(parent,Directory)):
		path = os.path.join(parent.name,path)
		parent = parent.eContainer()
	return path #2: omits the ./ at the beginning

@PathElementA.behavior
def actualPathAbs(self):
	path = self.name
	parent = self.eContainer()
	while(isinstance(parent,Directory)):
		path = os.path.join(parent.name,path)
		if(isinstance(parent, LocalDomainA)):
			path = os.path.join(parent.baseDirAbs,path)
		parent = parent.eContainer()
	return path #2: omits the ./ at the beginning

@PathElementA.behavior
def actualPathCwd(self):
	path = self.name
	parent = self.eContainer()
	while(isinstance(parent,Directory)):
		path = os.path.join(parent.name,path)
		if(isinstance(parent, LocalDomainA)):
			path = os.path.join(parent.baseDir,path)
		parent = parent.eContainer()
	return path #2: omits the ./ at the beginning
		

'''

'''

DEFAULT_TIMEOUT = 1.0

class LocalDomain(LocalDomainA):
	
	def __init__(self,baseDir,timeout=DEFAULT_TIMEOUT,metaDir=['./.meta'],enableBackup=True,logPath='./log',logLimit=10):
		self.baseDir = baseDir
		self.baseDirAbs = os.path.join(os.getcwd(),baseDir)
		self.baseDirUri = URI(baseDir) 
		self.resourceSet = ResourceSet()
		self.resourceObjectLUT = {} #a look-up table between eResources and ModelResource objects
		self.defaultTimeout = timeout
		self.metaDir = metaDir #the directory that contains model definitions
		#logging
		self.logPath = logPath
		self.logLimit = logLimit #the number of backups stored concurrently
		self.currentLogDir = None #is set later
		self.errorsLogger = None
		self.commandsLogger = None # is set later
		self.changesLogger = None # is set later
		#backup
		self.enableBackup = enableBackup
		self.currentBackupDir = None #is set later
		
		
		self.version = "v0.7"
		
		self.objectToIdLUT = {}
		self.idToObjectLUT = {}
		self.lastId = 0
		self.knownModelExtensions = ['ecore'] #this is used to identify the model files to be loaded. ecore files are known as model files by default.
		
		self.commandLUT = [
			self.__HelloCommand,
			self.__GoodbyeCommand,
			self.__SessionCommand,
			self.__StatusCommand,
			self.__ChangesCommand,
			self.__RetrieveCommand,
			self.__CreateCommand,
			self.__UpdateCommand,
			self.__CloneCommand,
			self.__UndoCommand,
			self.__CallCommand,
			self.__AsyncCallCommand,
			self.__CallStatusCommand,
			self.__AbortCallCommand,
			self.__CompoundCommand]
		
		#initialization
		self.id = self.__EncodeObjects(self)
		
		#initialize local EResource content method in order to support dynamic loading
		#EResource.python_class.isLoaded = lambda this: (this.uri in this.domain.resourceSet.resources)
		
		self.name = '.' #this is the root directories name
		
		self.__InitLogging()
		
		if(self.enableBackup):
			self.__Backup()
		
		self.__InitCommands()
		
		self.__LoadMetaModels()
		
		self.__LoadResourceTree()
				
		self.__InitActions()
		
		self.threadLock = threading.Lock()
		
					
	'''
	Must be called before any actions on the model
	default timeout is 1s
	'''	
		
	def Do(self,command : CommandA) -> ResultA:
		with self.threadLock: #if multiple threads enter, all but the first must wait here
			try:
				transaction = self.__BeginTransaction()
				result = self.DoAsTransaction(command,transaction)
			except Exception as e: #catch things going wrong in transaction closing and ending
				traceback.print_exc(file=sys.stdout)
				self.errorsLogger.error(e, exc_info=True)
				result = ErrorResult()
				result.commandType = command.type
				result.code = 9999
				result.message = '%s'%(str(e))
			finally:
				self.__EndTransaction(transaction)
		
		return result
	
	def DoAsTransaction(self,command : CommandA, transaction : int) -> ResultA:
		self.commandsLogger.info(CommandParser.CommandToString(command))
		result = self.__ExecuteCommand(command,transaction)
		self.commandsLogger.info(ResultParser.ResultToString(result))
		return result
	
	def Call(self,actionCall : ActionCall, transaction : Transaction) -> ListValue:
		#returnValues = ListValue()
		args = self.__ResolveHistory(actionCall.args, transaction)	
		if('load-resource' == actionCall.action):
			if(len(args.v)!=1 or args.v[0].type!=ValueTypesE.OBJECTREF):
				raise EoqActionError(10,'The number of args for action load-resource must be 1 and the type must be FileResource.')
			fileResource = self.__DecodeSingleObject(args.v[0])
			self.__LoadResource(fileResource,transaction)
			returnValues = self.__EncodeSingleObject(fileResource)
		elif('save-resource' == actionCall.action):
			if(len(args.v)!=1 or args.v[0].type!=ValueTypesE.OBJECTREF):
				raise EoqActionError(10,'The number of args for action save-resource must be 1 and the type must be FileResource.')
			fileResource = self.__DecodeSingleObject(args.v[0])
			self.__SaveResource(fileResource,transaction)
			returnValues = self.__EncodeSingleObject(fileResource)
		else:
			raise EoqActionError(99,'Action with name %s is unknown for the handler localDomain'%(actionCall.action))
		actionCall.returnValues = returnValues
		return

	def AsyncCall(self, actionCall : ActionCall, transaction : Transaction):
		raise EoqGeneralError(44,'AyncCall: LocalDomain does not support asynchronous actions.')

	def CallStatus(self, actionCall : ActionCall, transaction : Transaction):
		raise EoqGeneralError(45,'CallStatus: LocalDomain does not support asynchronous actions.')

	def AbortCall(self, actionCall : ActionCall, transaction : Transaction):
		raise EoqGeneralError(46,'AbortCall: LocalDomain does not support asynchronous actions.')
	
	#PRIVATE METHODS
		
	def __ExecuteCommand(self,command : CommandA, transaction : int):
		result = None
		#resolve any history already here?
		try:
			if(0 <= command.type.value and command.type.value < len(self.commandLUT)):
				result = self.commandLUT[command.type.value](command,transaction)
			else:
				raise EoqGeneralError(99,'Unknown command type: %d'%(command.type.value))

		#Error handling
		except EoqError as e:
			cmdStr = CommandParser.CommandToString(command)
			print('ERROR EXECUTING COMMAND:%s'%(cmdStr))
			traceback.print_exc(file=sys.stdout)
			self.errorsLogger.exception('ERROR EXECUTING COMMAND:%s'%(cmdStr))
			#self.errorsLogger.error(e, exc_info=True)
			result = ErrorResult()
			result.code = e.type + e.code
			result.message = '%s (in "%s")'%(e.message,cmdStr)
		except Exception as e:
			cmdStr = CommandParser.CommandToString(command)
			print('ERROR EXECUTING COMMAND:%s'%(cmdStr))
			traceback.print_exc(file=sys.stdout)
			self.errorsLogger.exception('ERROR EXECUTING COMMAND:%s'%(cmdStr))
			#self.errorsLogger.error(e, exc_info=True)
			result = ErrorResult()
			result.code = 9800
			result.message = '%s (in "%s")'%(str(e),cmdStr)
			
		result.commandType = command.type
		result.transactionId = transaction.id
		return result
	
	def __HelloCommand(self,command : HelloCommand,transaction : Transaction)->ResultA:
		result = HelloResult()
		#TODO: add return types
		return result
	
	def __GoodbyeCommand(self,command : GoodbyeCommand,transaction : Transaction)->ResultA:
		result = GoodbyeResult()
		return result
	
	def __SessionCommand(self,command : SessionCommand,transaction : Transaction)->ResultA:
		result = SessionResult()
		#TODO: add return types
		return result
	
	def __StatusCommand(self,command : StatusCommand,transaction : Transaction)->ResultA:
		result = StatusResult()
		#TODO: add return types
		return result
	
	def __ChangesCommand(self,command : ChangesCommand,transaction : Transaction)->ResultA:
		result = ChangesResult()
		earliestChangeId = self.__Resolve(command.earliestChangeId, IntValue, 'earliestChangeId', transaction).v
		requestedChanges = self.changes[earliestChangeId:]
		result.changes = ListValue()
		for requestedChange in requestedChanges:
			changeRecord = ListValue()
			changeRecord.v.add(IntValue(v=requestedChange.changeId))
			changeRecord.v.add(IntValue(v=requestedChange.sourceTransactionId))
			changeRecord.v.add(ValueParser.CloneValue(requestedChange.target))
			changeRecord.v.add(StringValue(v=QueryParser.QueryToString(requestedChange.query)))
			changeRecord.v.add(ValueParser.CloneValue(requestedChange.newValue))
			changeRecord.v.add(ValueParser.CloneValue(requestedChange.oldValue))
			result.changes.v.add(changeRecord)		
		return result
	
	def __RetrieveCommand(self,command : RetrieveCommand,transaction : Transaction)->ResultA:
		returnValue = self.__Retrieve(command.target, command.query,transaction)
		result = RetrieveResult()
		result.value = returnValue
		return result
	
	def __CreateCommand(self,command : CreateCommand,transaction : Transaction)->ResultA:
		
		packageNsUri = self.__Resolve(command.packageNsUri, StringValue, 'packageNsUri', transaction).v
		className = self.__Resolve(command.className, StringValue, 'className', transaction).v
		n = self.__Resolve(command.n, IntValue, 'n', transaction).v
		
		returnValue = self.__Create(packageNsUri, className, n,transaction)
		result = CreateResult()
		result.value = returnValue
		return result
	
	def __UpdateCommand(self,command : UpdateCommand,transaction : Transaction)->ResultA:
		returnValue = self.__Update(command.target, command.query, command.value, transaction)
		result = UpdateResult()
		result.target = returnValue
		return result
	
	def __CloneCommand(self,command : CloneCommand,transaction : Transaction)->ResultA:
		returnValue = self.__Clone(command.target, command.mode, transaction)
		result = CloneResult()
		result.value = returnValue
		return result
	
	def __UndoCommand(self,command,transaction : Transaction):
		pass
			
	def __CallCommand(self,command : CallCommand,transaction : Transaction)->ResultA:
		#initialize action call
		actionCall = self.__NewActionCall(command,transaction)
		self.actionCalls.add(actionCall)
		#execute action
		actionCall.handler.Call(actionCall,transaction)
		#build return value
		result = CallResult(callId=actionCall.callId,returnValues=actionCall.returnValues)
		return result
	
	def __AsyncCallCommand(self,command : AsyncCallCommand,transaction : Transaction)->ResultA:
		#initialize action call
		actionCall = self.__NewActionCall(command,transaction)
		#execute action
		actionCall.handler.AsyncCall(actionCall,transaction)
		#build return value
		result = AsyncCallResult(callId=actionCall.callId)
		return result
	
	def __CallStatusCommand(self,command : CallStatusCommand,transaction : Transaction)->ResultA:
		result = CallStatusResult()
		result.result = EmptyValue()
		#TODO: add return types
		return result
	
	def __AbortCallCommand(self,command : AbortCallCommand,transaction : Transaction)->ResultA:
		callId = self.__Resolve(command.callId, IntValue, 'callId', transaction).v
		if(callId<0 or len(self.actionCalls)<=callId):
			raise EoqGeneralError(23,'CallId %d is unknown'%(callId))
		actionCall = self.actionCalls[callId]
		actionCall.handler.AbortCall(actionCall,transaction)
		result = AbortCallResult()
		#TODO: add return types
		return result
	
	def __CompoundCommand(self,command : CompoundCommand,transaction : Transaction)->ResultA:
		result = CompoundResult()
		result.transactionId = transaction.id
		for subcommand in command.commands:
			subresult = self.__ExecuteCommand(subcommand,transaction)
			result.results.add(subresult)
			if(subresult.type == ResultTypesE.ERROR):
				result.type = ResultTypesE.COMPOUND_ERROR
				break #stop if one command fails
		return result
	
	def __NewActionCall(self,command,transaction : Transaction):
		action = self.__Resolve(command.action, StringValue, 'action', transaction).v
		actionHandler = self.__GetActionHandler(action)
		callId = self.callCount
		#resolve history arguments
		args = ValueParser.CloneValue(command.args)
		args = self.__ResolveHistory(args, transaction)	
		self.callCount +=1
		actionCall = ActionCall(callId=callId,action=action,args=args,handler=actionHandler)
		self.actionCalls.add(actionCall)
		return actionCall
		
	def __BeginTransaction(self,timeout=DEFAULT_TIMEOUT):
		if(timeout==DEFAULT_TIMEOUT):
			timeout = self.defaultTimeout #inherit the timeout from the class definition
		currentTime = time.perf_counter()
		if(self.currentTransaction):
			#check if the ongoing transaction is still in time
			if self.currentTransaction.deadline < currentTime:
				self.currentTransaction.wasTimedout = True
				self.currentTransaction.deadline = currentTime
			else:
				raise EoqTransactionError(1,'Cannot start transactionRecord another transaction is ongoing')
		#create a new transaction
		transactionId = self.transactionCount
		transaction = Transaction(id=transactionId,starttime=currentTime,maxDuration=timeout,deadline=(currentTime+timeout))
		self.transactions.add(transaction)
		self.transactionCount += 1
		
		self.currentTransaction = transaction
		return transaction
		
	def __EndTransaction(self,transaction):
		currentTime = time.perf_counter()
		transaction.wasEnded = True
		transaction.endtime = currentTime
		self.currentTransaction = None
		
	def __ValidateTransaction(self,transaction):
		currentTime = time.perf_counter()
		if self.currentTransaction == transaction:
			#check if the transaction has still time
			if transaction.deadline > currentTime:
				return transaction #everything seems ok, the transaction can go on
			else:
				#transaction.Timeout(currentTime)
				transaction.wasTimedout = True
				transaction.deadline = currentTime
				self.currentTransaction = None
				raise EoqTransactionError(2,'Transaction timeout. The requested transaction time of %f s is over since it is now %f s after begin. Quit here.'%(transaction.maxDuration,currentTime-transaction.starttime))
		else:
			if(transaction.wasEnded):
				raise EoqTransactionError(3,'Your transaction is not active any more. It was closed %f s ago.'%(currentTime-transaction.endtime))
			if(transaction.wasTimedout):
				raise EoqTransactionError(4,'Your transaction is not active any more. It timed out %f s ago.'%(currentTime-transaction.deadline))
			else:
				raise EoqTransactionError(5,'Serious error. Your transaction is not active any more. But reason is not clear.')
		return transaction
		
	def __Resolve(self,value : Value, expectedType : object, name : str, transaction : Transaction):
		resolvedValue = self.__ResolveHistory(value, transaction)
		if(not isinstance(resolvedValue,expectedType)):
			raise EoqValueError(33,'Wrong value type. Got value type %s for parameter %s, but expected type %s'%(type(resolvedValue).__name__,name,expectedType.__name__),value)
		return resolvedValue
		
	'''
	examples:
	- Retrieve(None,':architecture.ecore') #returns the root of resource 'architecture.ecore
	'''
	def __Retrieve(self,targetIds,query,transaction):
		transaction = self.__ValidateTransaction(transaction)
		
		hTargetIds = self.__ResolveHistory(targetIds,transaction)
		pTargetIds = ValueParser.ValueToPython(hTargetIds)
		targetObjs = self.__DecodeObjects(pTargetIds)
		#check the input type
		if(query.sourceClass):
			self.__ValidateSourceClass(targetObjs,query.sourceClass)
			
		
		#resultObjs = self.__RetrieveSegments(targetObjs,query.segments,transaction)
		resultObjs = self.__ResolveSegmentsRecursive(targetObjs,query.segments,0,1,transaction)
		
		
		#consider multiplicity
		if(query.returnMultiplicity):
			resultObjs = self.__ApplyReturnMultiplicity(resultObjs,query.returnMultiplicity)
		

		resultIds = self.__EncodeObjects(resultObjs)
		values = ValueParser.PythonToValue(resultIds)
		
		#store result in history	
		historyEntry = ValueParser.CloneValue(values) #must be cloned since it can be modified from outside otherwise
		self.__AddToHistory(historyEntry,transaction)
		
		return values
	
	def __Update(self,targetIds,query,value,transaction):
		transaction = self.__ValidateTransaction(transaction)
		hTargetIds = self.__ResolveHistory(targetIds,transaction)
		pTargetIds = ValueParser.ValueToPython(hTargetIds)
		targetObjs = self.__DecodeObjects(pTargetIds)
		hvalue = self.__ResolveHistory(value,transaction)
		pvalue = ValueParser.ValueToPython(hvalue)
		pvalue = self.__DecodeObjects(pvalue)

		nSegments = len(query.segments)
		lastSegment = None
		if(1==nSegments):
			updateObjs = targetObjs
			lastSegment = query.segments[0]
		elif(1<nSegments):
			lastSegment = query.segments[-1]
			#updateObjs = self.__RetrieveSegments(targetObjs, query.segments[:-1],transaction)
			updateObjs = self.__ResolveSegmentsRecursive(targetObjs, query.segments[:-1],0,1,transaction)
		
		#retrieve the actual values for the purpose of logging this transaction
		pOldValue = None
		if(isinstance(updateObjs,Resource)):
			pOldValue = []
		else:
			#pOldValue = self.__RetrieveSegments(updateObjs,[lastSegment],transaction)
			pOldValue = self.__ResolveSegmentsRecursive(updateObjs,[lastSegment],0,1,transaction)
		pOldValueNoObjects = self.__EncodeObjects(pOldValue)
		oldValue = ValueParser.PythonToValue(pOldValueNoObjects)
		
		#Carry out actual update
		self.__UpdateSegments(updateObjs,lastSegment,pvalue,transaction)	
		
		#get the new value in the new encoding (History values must be eliminated by real objects
		pNewValueNoObjects = self.__EncodeObjects(pvalue)
		newValue = ValueParser.PythonToValue(pNewValueNoObjects)
				
		#return the changed elements
		changedObjs = updateObjs
		changedObjsNoObject = self.__EncodeObjects(changedObjs)
		
		effectedObjects = ValueParser.PythonToValue(changedObjsNoObject)
		
		#store the changes in the record
		changedTargets = ValueParser.CloneValue(effectedObjects) #objects must be cloned, because they become alread member of the returned results
		lastSegmentOnlyQuery = Query()
		lastSegmentOnlyQuery.segments.append(lastSegment) #this line might change the initial query object.
		self.__AddToChangeRecord(changedTargets, lastSegmentOnlyQuery, newValue, oldValue, transaction)
		
		return effectedObjects
	
	def __Create(self,packageNsUri,className,numberOfInstances,transaction):
		transaction = self.__ValidateTransaction(transaction)
		#Create the new EObject
		obj = None
		try:
			pack = self.resourceSet.metamodel_registry[packageNsUri]
		except KeyError:
			raise EoqCreateError(5,'ERROR(NS): No package with namespace %s is known.'%(packageNsUri),packageNsUri,className)
		clazz = pack.getEClassifier(className)
		if(None is clazz): 
			raise EoqCreateError(6,'ERROR(class): No class named %s in package %s.'%(className,packageNsUri),packageNsUri,className)
		try:
			if(1==numberOfInstances):
				obj = clazz()
			elif(1<numberOfInstances):
				obj = []
				for i in range(numberOfInstances):
					obj.append(clazz())
		except TypeError as e:
			raise EoqCreateError(7,'ERROR(instantiation): %s.'%(e),packageNsUri,className)
		
		
		
		objectId = self.__EncodeObjects(obj)
		
		value = ValueParser.PythonToValue(objectId)
		#store result in history	
		historyEntry = ValueParser.CloneValue(value) #must be cloned since it can be modified from outside otherwise
		self.__AddToHistory(historyEntry,transaction)
		
		return value
	
	def __Clone(self,targetIds,mode,transaction):
		transaction = self.__ValidateTransaction(transaction)
		
		hTargetIds = self.__ResolveHistory(targetIds,transaction)
		pTargetIds = ValueParser.ValueToPython(hTargetIds)
		targetObjs = self.__DecodeObjects(pTargetIds)
		
		
		#resultObjs = self.__RetrieveSegments(targetObjs,query.segments,transaction)
		resultObjs = self.__CloneRecursive(targetObjs,mode,transaction)
		
		resultIds = self.__EncodeObjects(resultObjs)
		values = ValueParser.PythonToValue(resultIds)
		
		#store result in history	
		historyEntry = ValueParser.CloneValue(values) #must be cloned since it can be modified from outside otherwise
		self.__AddToHistory(historyEntry,transaction)
		
		return values
	
	def __CloneRecursive(self,targetObjs,mode,transaction):
		value = None
		if(isinstance(targetObjs, list)):
			value = []
			for targetObj in targetObjs:
				value.append(self.__CloneRecursive(targetObj, mode, transaction))
		else:
			if(CloneModesE.CLASS==mode):
				value = self.__ECloneClass(targetObjs)
			elif(CloneModesE.ATTRIBUTES==mode):
				value = self.__ECloneAttributes(targetObjs)
			elif(CloneModesE.FULL==mode):
				value = self.__ECloneFull(targetObjs)
			elif(CloneModesE.DEEP==mode):
				value = self.__ECloneDeep(targetObjs)
		return value
	
	def __AddToHistory(self,value,transaction):
		transaction.history.add(value) 
		
	def __ResolveHistory(self,value,transaction):
		#resolvedValue = value
		if(value.type == ValueTypesE.HISTORYREF):
			slot = value.v
			nHistory = len(transaction.history)
			if(slot<nHistory):
				if(slot<0):
					slot = nHistory+slot #slot is negative
				historyValue = transaction.history[slot]
				value = ValueParser.CloneValue(historyValue)
			else:
				raise EoqIdError(9,'Error history $%s does not exist. The current history length for this transaction is %d'%(slot,len(transaction.history)),slot)
		elif(value.type == ValueTypesE.LIST):
			nSubvalues = len(value.v)
			for i in range(nSubvalues):
				value.v[i] = self.__ResolveHistory(value.v[i], transaction)
		return value
		
	def __EncodeObjects(self,objs):
		if(isinstance(objs,list)):
			n = len(objs)
			for i in range(n):
				objs[i] = self.__EncodeObjects(objs[i])
		elif(isinstance(objs,EEnumLiteral)):
			objs = str(objs)
		elif(isinstance(objs,(EObject,type))):
			objs = self.__EncodeSingleObject(objs)
		return objs
	
	def __EncodeSingleObject(self,obj):
		idNb = 0
		try:
		##if obj in self.objectToIdLUT:
			idNb = self.objectToIdLUT[obj]
		except:
		#else:
			idNb = self.lastId
			self.objectToIdLUT[obj] = idNb
			self.idToObjectLUT[idNb] = obj
			self.lastId += 1 #root id starts now at 0 like all other indexes
		return ObjectRefValue(v=idNb)
	
	def __DecodeObjects(self,objIds):
		if(isinstance(objIds,list)):
			n = len(objIds)
			for i in range(n):
				objIds[i] = self.__DecodeObjects(objIds[i])
		elif(isinstance(objIds,ObjectRefValue)):
			objIds = self.__DecodeSingleObject(objIds)
		return objIds
	
	def __DecodeSingleObject(self,objId):
		obj = None
		idNb = objId.v
		try:
		#if idNb in self.idToObjectLUT:
			obj = self.idToObjectLUT[idNb]
		except:
		#else:
			raise EoqIdError(8,'Error id #%s is not a known object'%(idNb),idNb)
		return obj
		
	'''
	PRIVATE METHODS
	'''

	def __UpdateSegments(self,targetObjs,segment,value,transaction):
		if(isinstance(targetObjs,list)):
			nTargetObjs = len(targetObjs)
			isOneToOneAssignment = False
			if(isinstance(value,list)):
				nValues = len(value)
				if(nValues==nTargetObjs):
					isOneToOneAssignment = True #in this case each of the entries in values is assigned to exactly on target obj
			for i in range(nTargetObjs):
				targetObj = targetObjs[i]
				v = value[i] if isOneToOneAssignment else value		
				if(None == v):
					pass	
				if(isinstance(targetObj,list)):
					self.__UpdateSegments(targetObj,segment,v,transaction) #recursion
				else:
					self.__UpdateSegment(targetObj,segment,v,transaction)
		else:
			self.__UpdateSegment(targetObjs,segment,value,transaction)
		return
	
	def __UpdateSegment(self,targetObj,segment,value,transaction):
		if(segment.selector):
			raise EoqUpdateError(10,"ERROR: Set operation can not have a selector",value)
			return []
		if(segment.depth):
			raise EoqUpdateError(10,"ERROR: Set operation can not have a depth",value)
			return []
		if(isinstance(segment,PathSegment)):
			#for o in objs:
			self.__SetReferenceOrAttribute(targetObj,segment,value,transaction)
			#resource = targetObj.eResource
			#if(None != resource):
				#self.resourceDirty[resource] = True
		elif(isinstance(segment,ResourceSegment)):
			destination = self.__RetrieveResourceSegment(segment.identifier)
			#get the resource, or load it and clear existing elements if required.
			resource = None
			resourceContents = None
			
			if(isinstance(destination,Resource)): #resource is empty
				resource = destination
				resourceContents = []
			elif(isinstance(destination,list)): #resource has already a obj element, which is now replaced
				resource = destination[0].eResource
				resourceContents = destination
				
			#check for index
			if(segment.index):
				nResourceContents = len(resourceContents)
				if(segment.index.type == IndexTypesE.NUMBER):
					if(isinstance(value,list)):
						raise EoqUpdateError(11,"When setting a specific resource %d index, must be an EObject, not a list."%(segment.index.value),value)
					if(segment.index.value > nResourceContents):
						raise EoqUpdateError(12,"Resource has only %d elements. Cannot set index %"%(nResourceContents,segment.index.value),value)
					#because the resource array can not be edited directly, the elements that shall persist must be collected and later readded
					removedResources = [] 
					for i in range(nResourceContents):
						if(i>=segment.index.value):
							resource.remove(resourceContents[i])
						if(i>segment.index.value):
							removedResources.append(resourceContents[i])
					#add the new entry and all old ones
					if(isinstance(value,EObject)):
						resource.append(value)
					elif(None==value):
						pass # do nothing, which means the element is removed
					for c in removedResources:
						resource.append(c)
				if(segment.index.type == IndexTypesE.ADD):
					if(isinstance(value,list)):
						raise EoqUpdateError(13,"When setting a specific resource %d index, must be an EObject, not a list."%(segment.index.value),value)
					resource.append(value)	
				if(segment.index.type == IndexTypesE.REMOVE):
					if(isinstance(value,list)):
						raise EoqUpdateError(14,"When setting a specific resource %d index, must be an EObject, not a list."%(segment.index.value),value)
					resource.remove(value)
			else: #without index
				#delete all existing
				for c in resourceContents:
					resource.remove(c)
				#add new ones
				if(isinstance(value,list)):
					for v in value:
						if(not isinstance(v,EObject)):
							raise EoqUpdateError(15,"Could not set resource because value must be EObject or list of EObjects, but got %s"%(type(v)),value)
						resource.append(v)
				elif(isinstance(value,EObject)):
					resource.append(value)
				elif(value==None):
					pass #do nothing resource stays empty and can be deleted		
				else:
					raise EoqUpdateError(16,"Could not set resource because value must be EObject or list of EObjects, but got %s"%(type(value)),value)
			#self.resourceDirty[resource] = True
			return

	def __SetReferenceOrAttribute(self,targetObj,segment,value,transaction):
		identifier = segment.identifier
		
		feature = targetObj.eClass.findEStructuralFeature(identifier)
		if(None==feature):
			raise EoqUpdateError(30,"ERROR: Object of class %s has no feature named %s"%(targetObj.eClass,identifier),value)
		
		if(feature.many):
			if(segment.index):
				index = segment.index
				if(index.type == IndexTypesE.NUMBER):
					i = index.value; 
					if(i<0 or (feature.upperBound > 0 and i>= feature.upperBound)):
						raise EoqUpdateError(31,"Maximum number of elements is %d. Index %d is iligeal"%(feature.upperBound,i),value)
					if(None == value):
						valueAtIndex = targetObj.eGet(feature)[i] 
						targetObj.eGet(feature).remove(valueAtIndex)
					else:
						targetObj.eGet(feature)[i] = value 
				elif(index.type == IndexTypesE.RANGE):
					#TODO implement
					raise EoqUpdateError(32,"WARNING: Range indexing is currently not implemented",value)
				elif(index.type == IndexTypesE.ADD):
					if(isinstance(value,list)):
						for v in value:
							targetObj.eGet(feature).add(v)
					else:
						targetObj.eGet(feature).add(value)
				elif(index.type == IndexTypesE.REMOVE):
					if(isinstance(value,list)):
						for v in value:
							targetObj.eGet(feature).remove(v)
					else:
						targetObj.eGet(feature).remove(value)
				else:
					#Should never go here
					raise EoqUpdateError(33,"ERROR: Unsupported index type %s."%(type(index)),value)
			else: #no index
				targetObj.eGet(feature).clear()
				if(None != value):
					targetObj.eGet(feature).add(value)
		else: #not many
			if(segment.index):
				index = segment.index
				if(index.type == IndexTypesE.NUMBER):
					i = index.value-1; #because Python indexes beginning at 0
					# the only accepted index is 0
					if(i!=0):
						raise EoqUpdateError(40,"Maximum number of elements is %d. Index %d is iligeal"%(feature.upperBound,i),value)
					else:
						targetObj.eSet(feature,value) 
				elif(index.type == IndexTypesE.RANGE):
					raise EoqUpdateError(41,"Range indexing is not applicable for single element references",value)
				elif(index.type == IndexTypesE.ADD):
					if(None == targetObj.eGet(feature)):
						targetObj.eSet(feature,value)
					else:
						raise EoqUpdateError(42,"Maximum number of elements is %d. Cannot attach a further element by [+]"%(feature.upperBound),value)
				elif(index.type == IndexTypesE.REMOVE):
					targetObj.eSet(feature,None)
			else: #no index
				targetObj.eSet(feature,value)
	
		modelResource = self.__GetContainingResourceObject(targetObj)
		if(modelResource):
			self.__SetResourceDirty(modelResource, transaction)
	
	def __ResolveSegmentsLinear(self,targetObjs,segments,currentSegment,currentDepth,transaction):
		resultObjs = None
		nSegments = len(segments)
		if(currentSegment < nSegments and targetObjs):
			childObjs = None
			segment = segments[currentSegment]
			if(segment.identifierType == ListOpSegmentTypesE.from_string('SIZE')):
				if(isinstance(targetObjs,list)):
					childObjs = len(targetObjs)
				else:
					childObjs = 1;
			elif(segment.identifierType == ListOpSegmentTypesE.from_string('FLATTEN')):
				pass
			#continue recursion
			resultObjs = self.__ResolveSegmentsRecursive(childObjs,segments,currentSegment+1,1,transaction) #start path recursion	
		else: #recursion is over
			resultObjs = targetObjs
		return resultObjs
		
	def __ResolveSegmentsRecursive(self,targetObjs,segments,currentSegment,currentDepth,transaction):
		resultObjs = None
		#parentObjs = targetObjs
		nSegments = len(segments)
		if(currentSegment < nSegments and targetObjs): #non exiting objects are as null):
			segment = segments[currentSegment]
			
			#evaluate the current segment
			if(SegmentProcessingTypesE.LINEAR == segment.processingType):
				resultObjs = self.__ResolveSegmentsLinear(targetObjs,segments,currentSegment,1,transaction)
			else:
				if(isinstance(targetObjs,list)):
					resultObjs = []
					for targetObj in targetObjs:
						
						resultObj = self.__ResolveSegmentsRecursive(targetObj,segments,currentSegment,1,transaction)
						resultObjs.append(resultObj)
				else: #Could check for none here, but prefere that the user gets and error if the following is an unset object
					resultObj = self.__ResolveSegment(targetObjs,segment,transaction)
					#apply index
					if(segment.index):
						resultObj = self.__ApplyIndex(resultObj, segment.index)
					
					childObjs = self.__ResolveSegmentsRecursive(resultObj,segments,currentSegment+1,1,transaction) #start path recursion
					
					#apply depth (Seems realy complex think about disabling depth for deterministic applications
					if(segment.depth and segment.depth.value != currentDepth):
						if(isinstance(resultObj,list)):
							for i in range(len(resultObj)):
								recursiveChildObj = self.__ResolveSegmentsRecursive(resultObj[i],segments,currentSegment,currentDepth+1,transaction) #start depth recursion
								if(recursiveChildObj):
									if(not isinstance(childObjs,list)):
										childObjs[i] = [childObjs[i]]
									if(isinstance(recursiveChildObj,list)):
										childObjs[i].extend(recursiveChildObj)
									else:
										childObjs[i].append(recursiveChildObj)
						#merge results
						elif(resultObj):
							recursiveChildObj = self.__ResolveSegmentsRecursive(resultObj,segments,currentSegment,currentDepth+1,transaction) #start depth recursion
							if(recursiveChildObj):
								if(not isinstance(childObjs,list)):
									childObjs = [childObjs]
								if(isinstance(recursiveChildObj,list)):
									childObjs.extend(recursiveChildObj)
								else:
									childObjs.append(recursiveChildObj)
						
					resultObjs = childObjs	
		else: #recursion is over
			resultObjs = targetObjs
		return resultObjs
		
	def __ResolveSegment(self,targetObj,segment,transaction):
		resultObjs = None
		
		#selector
		selectorFunctor = self.__CreateSelectorFunctor(segment,transaction)
		
		#Eval individual segment types
		if(segment.type == SegmentTypesE.PATH):
			resultObjs = self.__RetrievePathSegment(targetObj,segment.identifier,selectorFunctor)
		elif(segment.type == SegmentTypesE.CLAZZ):
			resultObjs = self.__RetrieveClassSegment(targetObj,segment.identifier,selectorFunctor)
		elif(segment.type == SegmentTypesE.INSTANCE):
			resultObjs = self.__RetrieveInstanceOfSegment(targetObj,segment.identifier,selectorFunctor)
# 		elif(segment.type == SegmentTypesE.ID):
# 			childs = self.__RetrieveIdSegment(targetObj,segment.identifier)
		elif(segment.type == SegmentTypesE.META):
			resultObjs = self.__RetrieveMetaSegment(targetObj,segment,selectorFunctor)
		elif(segment.type == SegmentTypesE.RESOURCE):
			resultObjs = self.__RetrieveResourceSegment(segment.identifier)
		return resultObjs
	
	def __CreateSelectorFunctor(self,segment,transaction):
		
		#consider the selector
		selectorFunctor = lambda o: True
		
		if(segment.selector):
			#initialization
			comperatorFunctor = lambda a,b:False
			selectFunctor = lambda o:False
			
			#handle selector
			selector = segment.selector
			selectFunctor  = lambda o: o.eGet(selector.name)
			
			#handle operation
			operator = selector.operator
			if(operator.type == OperatorTypesE.EQUAL):
				comperatorFunctor = lambda a,b: a==b
			elif(operator.type == OperatorTypesE.NOTEQUAL):
				comperatorFunctor = lambda a,b: a!=b
			elif(operator.type == OperatorTypesE.GREATER):
				comperatorFunctor = lambda a,b: a>b
			elif(operator.type == OperatorTypesE.LESS):
				comperatorFunctor = lambda a,b: a<b
			
			#handle value type
			valueObj = selector.value
			if(valueObj.type == ValueTypesE.INT):
				value = valueObj.v
			elif(valueObj.type == ValueTypesE.FLOAT):
				value = valueObj.v
			elif(valueObj.type == ValueTypesE.BOOL):
				value = valueObj.v
			elif(valueObj.type == ValueTypesE.STRING):
				value = valueObj.v
			elif(valueObj.type == ValueTypesE.OBJECTREF):
				value = self.__DecodeSingleObject(valueObj)
			elif(valueObj.type == ValueTypesE.HISTORYREF):
				value = self.__ResolveHistory(valueObj,transaction)
			elif(valueObj.type == ValueTypesE.EMPTY):
				value = None
			
			selectorFunctor  = lambda o: self.__DoesObjectMatchSelector(o,comperatorFunctor,selectFunctor,value)
		
		return selectorFunctor
	
	def __DoesObjectMatchSelector(self,obj,comperatorFunctor,selectFunctor,value):
		a = selectFunctor(obj)
		if(isinstance(a,EOrderedSet)):
			for memberOfA in a:
				if(comperatorFunctor(memberOfA,value)):
					return True
			return False #if no comparison was successful the result is false
		else:
			return comperatorFunctor(a,value)

	
	def __RetrievePathSegment(self,targetObj,refname,selectorFunctor):
		resultObjs = None
		#children = targetObj.eGet(refname)
		children = targetObj.__getattribute__(refname); #hack for more speed
		if(isinstance(children,types.MethodType)):
			#children = targetObj.eGet(refname)() #imitate a method call
			children = targetObj.__getattribute__(refname)() #imitate a method call
		if(isinstance(children,(EOrderedSet,list,set))):
			resultObjs = []
			for child in children:
				if(selectorFunctor(child)):
					resultObjs.append(self.__ResolveProxyAndBuildInClasses(child))
		elif(selectorFunctor(children)):
			resultObjs = self.__ResolveProxyAndBuildInClasses(children)
		return resultObjs
	
	def __ResolveProxyAndBuildInClasses(self,obj): 
		if(isinstance(obj, EProxy)):
			obj.force_resolve()
			obj = obj._wrapped #remove the outer proxy
		if(isinstance(obj, (MetaEClass,type,types.ModuleType))): #this is necessary to mask compiled model instances
			obj = obj.eClass
		return obj
	
	def __ApplyIndex(self,objs,segmentIndex):
		if(segmentIndex.type in [IndexTypesE.NUMBER,IndexTypesE.RANGE] and isinstance(objs,list)):
			nObjs = len(objs)
			if(segmentIndex.type == IndexTypesE.NUMBER):
				if(segmentIndex.value >= nObjs):
					raise EoqRetrieveError(10,'Index value %d is greater than result size of %d'%(segmentIndex.value,nObjs))
				return objs[segmentIndex.value]
			elif(segmentIndex.type == IndexTypesE.RANGE):
				if(segmentIndex.upper >= nObjs) :
					raise EoqRetrieveError(11,'Upper range index %d is greater than the result size of %d'%(segmentIndex.upper,nObjs))
				return objs[segmentIndex.lower:segmentIndex.upper]
		else:
			return objs
	
	def __RetrieveClassSegment(self,targetObj,clsname,selectorFunctor):
		resultObjs = []

		if(self.__IsType(targetObj,clsname) and selectorFunctor(targetObj)):
			resultObjs.append(targetObj)
		contents = targetObj.eAllContents()
		if(contents):
			for c in contents:
				if(self.__IsType(c,clsname) and selectorFunctor(c)):
					resultObjs.append(c)		
		return resultObjs
	
	def __RetrieveInstanceOfSegment(self,targetObj,clsname,selectorFunctor):
		resultObjs = []
		if(self.__IsInstanceOf(targetObj,clsname) and selectorFunctor(targetObj)):
			resultObjs.append(targetObj)
		contents = targetObj.eAllContents()
		if(contents):
			for c in contents:
				if(self.__IsInstanceOf(c,clsname) and selectorFunctor(c)):
					resultObjs.append(c)		
		return resultObjs
	
# 	def __RetrieveIdSegment(self,objs,idstr):
# 		results = []
# 		for o in objs:
# 			r = []	
# 			results += r
# 		return r
	
	def __RetrieveMetaSegment(self,targetObj,segment,selectorFunctor):
		resultObjs = []
		if(segment.identifierType == MetaSegmentIdentifiersE.from_string('CONTAINER')):
			resultObjs = targetObj.eContainer()
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('INDEX')):
			f = targetObj.eContainmentFeature()
			if(None==f or f.upperBound == 1):
				resultObjs = 0 ##-1 denotes the index makes no sense for single object references
			else: #index must be determined by backwards search
				#this backward search seems not optimal
				containingIndex = 0 #Eoq indicies start at 1
				container = targetObj.eContainer()
				for sibling in container.eGet(f):
					containingIndex = containingIndex+1
					if(sibling==targetObj):
						resultObjs = containingIndex
						break;
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('TYPE')):
			resultObjs = targetObj.eClass.findEStructuralFeature(segment.identifier) #does not make sence since it only works for 
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('CLASS')):
			resultObjs = self.__ResolveProxyAndBuildInClasses(targetObj.eClass)
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('CONTENTS')):
			resultObjs = [x for x in targetObj.eContents if selectorFunctor(x)]
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('CLASSNAME')):
			resultObjs = targetObj.eClass.name
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('PACKAGE')):
			resultObjs = self.__ResolveProxyAndBuildInClasses(targetObj.eClass.eContainer())
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('PACKAGENAME')):
			resultObjs = targetObj.eClass.eContainer().name
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('RESOURCE')):
			resultObjs = self.__GetContainingResourceObject(targetObj)
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('RESOURCENAME')):
			resultObjs = ''
			for key,value in self.resourceSet.resources.items():
				if value==targetObj.eResource:
					resultObjs = self.__ResourcePathOnly(key)
					break;
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('CONTAININGFEATURE')):
			f = self.__ResolveProxyAndBuildInClasses(targetObj.eContainmentFeature())
			if(f):
				resultObjs = f
			else:
				resultObjs = None
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('CONTAININGFEATURENAME')):
			f = targetObj.eContainmentFeature()
			if(f):
				resultObjs = f.name
			else:
				resultObjs = ''
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('FEATURES')):
			resultObjs = [self.__ResolveProxyAndBuildInClasses(x) for x in targetObj.eClass.eAllStructuralFeatures() if selectorFunctor(x)]
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('FEATURENAMES')):
			resultObjs = [x.name for x in targetObj.eClass.eAllStructuralFeatures() if selectorFunctor(x)]
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('FEATUREVALUES')):
			features = [x.name for x in targetObj.eClass.eAllStructuralFeatures() if selectorFunctor(x)]
			resultObjs = [self.__OrderedSetEliminator(targetObj.eGet(a)) for a in features]
			for i in range(len(resultObjs)):
				if(isinstance(resultObjs[i], list)):
					for j in range(len(resultObjs[i])):
						resultObjs[i][j] = self.__ResolveProxyAndBuildInClasses(resultObjs[i][j])
				else:
					resultObjs[i] = self.__ResolveProxyAndBuildInClasses(resultObjs[i])
			#resultObjs = [[ self.__ResolveProxyAndBuildInClasses(x) for x in self.__OrderedSetEliminator(targetObj.eGet(a))] for a in features]
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('ATTRIBUTES')):
			resultObjs = [self.__ResolveProxyAndBuildInClasses(x) for x in targetObj.eClass.eAllAttributes() if selectorFunctor(x)]
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('ATTRIBUTENAMES')):
			resultObjs = [x.name for x in targetObj.eClass.eAllAttributes() if selectorFunctor(x)]
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('ATTRIBUTEVALUES')):
			attributes = [x for x in targetObj.eClass.eAllAttributes() if selectorFunctor(x) ]
			resultObjs = [targetObj.eGet(a) for a in attributes]
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('EALLREFERENCES')):
			resultObjs = [self.__ResolveProxyAndBuildInClasses(x) for x in targetObj.eAllReferences() if selectorFunctor(x)] 
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('REFERENCES')):
			resultObjs = [self.__ResolveProxyAndBuildInClasses(x) for x in targetObj.eClass.eAllReferences() if selectorFunctor(x)] 
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('REFERENCENAMES')):
			resultObjs = [x.name for x in targetObj.eClass.eAllReferences() if selectorFunctor(x)]
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('REFERENCEVALUES')):
			references = [x.name for x in targetObj.eClass.eAllReferences() if selectorFunctor(x)]
			resultObjs = [self.__OrderedSetEliminator(targetObj.eGet(a)) for a in references]
			for i in range(len(resultObjs)):
				if(isinstance(resultObjs[i], list)):
					for j in range(len(resultObjs[i])):
						resultObjs[i][j] = self.__ResolveProxyAndBuildInClasses(resultObjs[i][j])
				else:
					resultObjs[i] = self.__ResolveProxyAndBuildInClasses(resultObjs[i])
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('CONTAINMENTS')):
			resultObjs = [self.__ResolveProxyAndBuildInClasses(x) for x in targetObj.eClass.eAllReferences() if x.containment and selectorFunctor(x)] 
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('CONTAINMENTNAMES')):
			resultObjs = [x.name for x in targetObj.eClass.eAllReferences() if x.containment and selectorFunctor(x)] 
		elif(segment.identifierType == MetaSegmentIdentifiersE.from_string('CONTAINMENTVALUES')):
			containments = [x.name for x in targetObj.eClass.eAllReferences() if x.containment and selectorFunctor(x)]
			resultObjs = [self.__OrderedSetEliminator(targetObj.eGet(a)) for a in containments] 
		else:
			raise EoqRetrieveError(51,'Illegal meta segment %s for object'%(segment.identifier))
		return resultObjs
	
	def __SetResourceDirty(self,modelResource,transaction):
		if(not modelResource.isDirty):
			modelResource.isDirty = True
			self.__AddToChangeRecord(self.__EncodeSingleObject(modelResource), QueryParser.StringToQuery('/isDirty'), BoolValue(v=True), BoolValue(v=False), transaction)
	
	def __SetResourceClean(self,modelResource,transaction):
		if(modelResource.isDirty):
			modelResource.isDirty = False
			self.__AddToChangeRecord(self.__EncodeSingleObject(modelResource), QueryParser.StringToQuery('/isDirty'), BoolValue(v=False), BoolValue(v=True), transaction)
	
	def __SetResourceLoaded(self,modelResource,transaction):
		if(not modelResource.isLoaded):
			modelResource.isLoaded = True
			self.__AddToChangeRecord(self.__EncodeSingleObject(modelResource), QueryParser.StringToQuery('/isLoaded'), BoolValue(v=True), BoolValue(v=False), transaction)
	
	def __SetResourcePersistentPath(self,modelResource,newPath,transaction):
		if(modelResource.lastPersistentPath != newPath):
			oldPath = modelResource.lastPersistentPath
			modelResource.lastPersistentPath = newPath
			self.__AddToChangeRecord(self.__EncodeSingleObject(modelResource), QueryParser.StringToQuery('/lastPersistentPath'), StringValue(v=newPath), StringValue(v=oldPath) if oldPath else EmptyValue(), transaction)

	
	def __GetContainingResourceObject(self,obj):
		resourceObj = None
		eResource = obj.eResource
		if(eResource in self.resourceObjectLUT):
			resourceObj = self.resourceObjectLUT[eResource]
		else:
			resourceObj = self.__RecursiveResourceObjSearch(obj)
		return resourceObj
	
	def __RecursiveResourceObjSearch(self,obj):
		resourceObj = None
		if(isinstance(obj,FileResourceA)):
			resourceObj = obj
		else:
			container = obj.eContainer()
			if(container):
				resourceObj = self.__RecursiveResourceObjSearch(container)
		return resourceObj
	
	def __RetrieveResourceSegment(self,resourceName):
		resultObjs = None
		resourcePath = os.path.join(self.baseDir,resourceName)
		resourceUri = URI(resourcePath)
		if resourceUri.normalize() not in self.resourceSet.resources:
			if(os.path.isfile(resourcePath)):
				if('.ecore' in resourceName and 'Ecore.ecore' not in resourceName): #register a meta model automatically. 
					#Prevent that the Ecore definition is registered as a new meta-model. Since the Ecore meta model is implemented in 
					#pyecore already, this will create a mess or name ambiguities.
					gcmRoot = self.__LoadMetaModelResource(resourcePath)
				else: 
					gcmRoot = self.__LoadModelResource(resourcePath)
				resultObjs = gcmRoot
			else:
				resource = self.resourceSet.create_resource(resourcePath)
				resultObjs = resource
		else:
			resource = self.resourceSet.resources[resourceUri.normalize()]
			if(len(resource.contents)>0):
				resultObjs = list(resource.contents) #resource list is an object, so if it is modified later this makes problems. Therefore a copy must be created.
			else:
				resultObjs = resource
		return resultObjs
	
	'''
	UTILITY FUNCTIONS
	'''
	
	def __IsType(self,obj,classname):
		return (obj.eClass.name==classname)
	
	def __IsInstanceOf(self,obj,classname):
		result = self.__IsType(obj,classname)
		if(not result):
			for superType in obj.eClass.eAllSuperTypes():
				if(superType.name==classname):
					result = True
					break
			if('EObject'==classname):
				result = isinstance(obj,EObject)
		return result
	
	def __ResourcePathOnly(self,absoluteResourcePath):
		absoluteBaseDirPath = self.baseDirUri.normalize();
		#absoluteResourcePath = resourceUri.normalize() 
		
		if(absoluteBaseDirPath in absoluteResourcePath):
			resourcePath = absoluteResourcePath[len(absoluteBaseDirPath)+1:]
		else:
			resourcePath = absoluteResourcePath
		return resourcePath
	
	def __ScanForModelFiles(self):
		modelFiles = []
		for root, dirs, files in os.walk(self.baseDir, topdown=True):
			for file in files:
				for extension in self.knownModelExtensions:
					if(file.endswith('.%s'%(extension))):
						relativeRoot = os.path.relpath(root, self.baseDir)
						modelFiles += [os.path.join(relativeRoot,file)] #2: omits the ./ at the beginning
		return modelFiles
	
	def __ValidateSourceClass(self,root,sourceClass):
		isContextValid = False
		className = sourceClass.name
		if(isinstance(root,list)):
			for obj in root:
				self.__ValidateSourceClass(obj, sourceClass)
		else:
			if(root.eClass.name==className):
				isContextValid = True
			elif('EObject'==className):
				isContextValid = isinstance(root,EObject)
			else:
				for superType in root.eClass.eAllSuperTypes():
					if(superType.name==className):
						isContextValid = True
						break
			if(not isContextValid):
				raise EoqContextError(60,"Context should be instance of class %s, but got object of class %s"%(className,root.eClass.name))
		#return isContextValid
	
	def __ApplyReturnMultiplicity(self,r,returnMultiplicity):
		if(returnMultiplicity.type == ReturnMultiplicityTypeE.FLATTENED):
			r = self.__FlattenList(r)
		elif(returnMultiplicity.type == ReturnMultiplicityTypeE.FORCESINGLE):
			if(isinstance(r, list)):
				if(1==len(r)):
					r = r[0]
				elif(0==len(r)):
					r = None
				else:
					raise EoqContextError((70,"Expected multiplicity is %d but got list of %d elements")%(1,len(r)))
		elif(returnMultiplicity.type == ReturnMultiplicityTypeE.UNCHANGED):
			pass #needs no action
		return r
	
	def __LoadModelResource(self,resourceFile):
		resource = self.resourceSet.get_resource(resourceFile)
		return list(resource.contents) #make a copy of the list contents
	
	def __LoadMetaModelResource(self,resourceFile):
		modelContents = self.__LoadModelResource(resourceFile)
		gcmRoot = modelContents[0]
		self.resourceSet.metamodel_registry[gcmRoot.nsURI] = gcmRoot
		# register all possible subpackages
		for child in gcmRoot.eAllContents():
			if(isinstance(child,EPackage)):
				self.resourceSet.metamodel_registry[child.nsURI] = child
		#remember the file extension, which is the file name (without extension) of the model file
		metaModelFileName = os.path.basename(resourceFile)
		modelExtension = os.path.splitext(metaModelFileName)[0]
		self.knownModelExtensions += [modelExtension]
		#return the root
		return modelContents
	
	def __OrderedSetEliminator(self,obj):
		if(isinstance(obj,EOrderedSet)):
			return list(obj)
		else:
			return obj
		
	def __FlattenList(self,nestedList):
		flatList = None
		if(isinstance(nestedList,list)):
			flatList = []
			for l1 in nestedList:
				if(isinstance(l1,list)):
					flatList.extend(self.__FlattenList(l1))
				else:
					flatList.append(l1)
		else:
			flatList = nestedList
		return flatList
	
	def __AddToChangeRecord(self,target,query,newValue,oldValue,transaction):
		changeRecord = Change()
		changeRecord.target = target
		changeRecord.query = query
		changeRecord.newValue = newValue
		changeRecord.oldValue = oldValue
		changeRecord.sourceTransactionId = transaction.id
		changeRecord.changeId = self.changeCount
		self.changes.add(changeRecord)
		self.changeCount += 1
		self.changesLogger.info("%d: %s %s %s %s %d"%(self.changeCount,
													ValueParser.ValueToString(target),
													QueryParser.QueryToString(query),
													ValueParser.ValueToString(newValue),
													ValueParser.ValueToString(oldValue),
													transaction.id))
	
	def __CloneList(self,inList):
		clonedList = None
		if(isinstance(inList,list)):
			clonedList = list(inList)
			for i in range(len(clonedList)):
				clonedList[i] = self.__CloneList(clonedList[i])
		else:
			clonedList = inList
		return clonedList
	
	def __InitLogging(self):
		#create the log dir if it does not exist
		if(not os.path.isdir(self.logPath)):
			os.makedirs(self.logPath)
		#build log folder name from current date
		now = datetime.now() # current date and time
		logBasename = now.strftime("%Y-%m-%d_%H_%M_%S")
		n = 0
		#check if there is already a backup for that second
		while(os.path.isdir(os.path.join(self.logPath,"%s_%05d"%(logBasename,n)))):
			n+=1
		self.currentLogDir = os.path.join(self.logPath,"%s_%05d"%(logBasename,n))
		
		#remove old backups, if limit is exceeded
		existingLogDirs = []
		for name in os.listdir(self.logPath): 
			existingLogDir = os.path.join (self.logPath, name)   
			if os.path.isdir(existingLogDir):
				existingLogDirs.append(existingLogDir)
				
		nExistingLogDirs = len(existingLogDirs)
		nEntriesToDelete = max([nExistingLogDirs-self.logLimit+1,0]) #+1 for the newly created one
		if 0 < nEntriesToDelete:
			existingLogDirs = sorted(existingLogDirs)
			for i in range(nEntriesToDelete):
				shutil.rmtree(existingLogDirs[i])
			nExistingLogDirs -= nEntriesToDelete
		
		#on for the current log
		os.makedirs(self.currentLogDir)
		nExistingLogDirs += 1;
		
		#init error logger
		self.errorsLogger = logging.getLogger('error')
		fh = logging.FileHandler(os.path.join(self.currentLogDir,"errors.log"))
		fh.setLevel(logging.INFO)
		formatter = logging.Formatter('%(asctime)s - %(message)s')
		fh.setFormatter(formatter)
		self.errorsLogger.addHandler(fh)
		self.errorsLogger.setLevel(logging.INFO)
		
		#init command logger
		self.commandsLogger = logging.getLogger('commands')
		fh = logging.FileHandler(os.path.join(self.currentLogDir,"commands.log"))
		fh.setLevel(logging.INFO)
		formatter = logging.Formatter('%(asctime)s - %(message)s')
		fh.setFormatter(formatter)
		self.commandsLogger.addHandler(fh)
		self.commandsLogger.setLevel(logging.INFO)
		
		#init changes logger
		self.changesLogger = logging.getLogger('changes')
		fh = logging.FileHandler(os.path.join(self.currentLogDir,"changes.log"))
		fh.setLevel(logging.INFO)
		formatter = logging.Formatter('%(asctime)s - %(message)s')
		fh.setFormatter(formatter)
		self.changesLogger.addHandler(fh)
		self.changesLogger.setLevel(logging.INFO)
		
		print('Logging enabled! (path: %s, existing: %d, removed: %d, limit: %d)'%(self.currentLogDir,nExistingLogDirs,nEntriesToDelete,self.logLimit))
	
	
	def __Backup(self):
		self.currentBackupDir = os.path.join(self.currentLogDir,"backup")
		
		#os.makedirs(backupName)
		shutil.copytree(self.baseDir,self.currentBackupDir)
		
		print('Backup created! (path: %s)'%(self.currentBackupDir))
	
	
	def __InitCommands(self):
		#load all commands from the meta model and initialize the commands data objects within the domain, such that the information is accessible by queries
		#model.eClass.eClassifiers
		#dummyCommand = RetrieveCommand() #use this to get access to the meta model
		#for commandClass in CommandA.eClass.allInstances(None):
		for classifier in CommandA.eClass.ePackage.eClassifiers:
			if(isinstance(classifier,EClass) and (CommandA.eClass in classifier.eAllSuperTypes())):
				commandClass = classifier
				commandFeatures = commandClass.eStructuralFeatures
				commandName = commandFeatures[0].get_default_value().name #the first feature of each command should always be the type
				commandDescription = self.__GetEAnnotationValue(commandClass,'documentation')
				command = CommandInfo(name=commandName,description=commandDescription)		
				for i in range(1,len(commandFeatures)):
					commandFeature = commandFeatures[i]
					paramName = commandFeature.name 
					paramType = commandFeature.eType.eClass.name
					paramDescription = self.__GetEAnnotationValue(commandFeature,'documentation')
					paramExample = 'exampleParameter'
					if(isinstance(commandFeature, EAttribute) and commandFeature.get_default_value()):
						paramExample = str(commandFeature.get_default_value()) #could be also an integer
					commandParameter = CommandParameter(name=paramName,type=paramType,description=paramDescription,example=paramExample)
					
					command.parameters.add(commandParameter)
				
				#find matching command results
				for resultClassifier in CommandA.eClass.ePackage.eClassifiers:
					if(isinstance(resultClassifier,EClass) and (ResultA.eClass in resultClassifier.eAllSuperTypes())):
						commandResultClass = resultClassifier
						commandResultFeatures = commandResultClass.eStructuralFeatures
						commandResultCommandType = commandResultFeatures[1].get_default_value().name #the second feature of each command should always be the command type
						if(commandResultCommandType == commandName):
							for j in range(1,len(commandResultFeatures)):
								commandResultFeature = commandResultFeatures[j]
								commandResultParamName = commandResultFeature.name
								commandResultParamType = commandResultFeature.eType.eClass.name
								commandResultParamDescription = self.__GetEAnnotationValue(commandResultFeature,'documentation')
								commandResultParamExample = 'exampleResult';
								if(isinstance(commandResultFeature, EAttribute) and commandResultFeature.get_default_value()):
									commandResultParamExample = str(commandResultFeature.get_default_value()) #could be also an integer
								commandResultParameter = CommandParameter(name=commandResultParamName,type=commandResultParamType,description=commandResultParamDescription,example=commandResultParamExample)
								command.results.add(commandResultParameter)
				#finally add the command definition to the domain such that it can be queried
				self.commands.add(command)
		return
				
				
	def __GetEAnnotationValue(self,annotatedElement,key):
		for annotation in annotatedElement.eAnnotations:
			for entry in annotation.entries:
				if(entry.key == key):
					return entry.value
				
		return ''
		
			
	def __LoadMetaModels(self):
		#look for existing meta models and load them on them
		for mp in self.metaDir:
			metadir = mp
			if(not os.path.isabs(mp)):
				metadir = os.path.join(self.baseDir,mp)
			searchString = os.path.join(metadir+'/*.ecore')
			modeldefinitions = glob.glob(searchString)
			for md in modeldefinitions:
				self.__LoadMetaModelResource(md)
					
		#transfer the complete meta registry to the domain model
		for uri in self.resourceSet.metamodel_registry.keys():
			package = self.resourceSet.metamodel_registry[uri]
			if(isinstance(package,types.ModuleType)):
				#this happens for metamodels loaded by generated python code
				#in this case we must create a copy as a psydo EPackge
				package = package.eClass #this is how pyecore generates the models. It is not clear, why the class of this module is a package, but it seems to work
				metamodel = Metamodel(name=package.name,source=MetaModelSourcesE.BUILDIN,package=package)
				self.metamodels.add(metamodel)
			else:
				metamodel = Metamodel(name=package.name,source=MetaModelSourcesE.DYNAMIC,package=package)
				self.metamodels.add(metamodel)
		return
						
			
	def __LoadResourceTree(self):
		modelfiles = self.__ScanForModelFiles()
		
		for modelfile in modelfiles:
			head,tail = os.path.split(modelfile)
			directory = self.__GetOrCreateDir(head)
			modelResource = ModelResource(name=tail,lastPersistentPath=modelfile)
			#do not load the resource now, do it later on demand
			directory.resources.add(modelResource)
		return
			
	def __InitActions(self):
		#register resource related actions 
		action = ActionInfo(name='load-resource',handler=self,description='Loads model elements from file based resources.')
		action.parameters.add(ActionParameter(name='resource',type='EResource',description='EResource object which contains the file path of the resource to be loaded and which will contain the data afterwards.'))
		action.results.add(ActionParameter(name='resource',type='EResource',description='The resource object that was loaded.'))
		self.actions.add(action)
		
		action = ActionInfo(name='save-resource',handler=self,description='Saves model elements to file based resources.')
		action.parameters.add(ActionParameter(name='resource',type='EResource',description='EResource object which contains the file path of the resource to be saved and which contains the data to be saved.'))
		action.results.add(ActionParameter(name='resource',type='EResource',description='The resource object that was saved.'))
		self.actions.add(action)	
		return
	
	def __GetOrCreateDir(self,path):
		directory = self
		if(path and path != '.'): #only procced for non empty strings
			segments = path.split(os.path.sep)
			for segment in segments: 
				subdirexists = False
				for subdir in directory.subdirectories:
					if(subdir.name == segment):
						directory = subdir
						subdirexists = True
						break
				if(not subdirexists):
					newsubdir = Directory(name=segment)
					directory.subdirectories.add(newsubdir)
					directory = newsubdir
		return directory
	
	def __GetActionHandler(self,actionName):
		for action in self.actions:
			if(action.name == actionName):
				return action.handler
		raise EoqActionError(1,'Action with name %s is unknown.'%(actionName))
		return
	
	#ACTIONS
	def __LoadResource(self,modelResource,transaction):
		if(modelResource.isLoaded):
			return modelResource
		if(modelResource.type == FileResourceTypesE.MODEL):
			resource = self.resourceSet.get_resource(os.path.join(self.baseDir,modelResource.lastPersistentPath)) #neglegt any path changes here, since they manifest on safe
			#self.DoAsTransaction(CommandParser.UpdateCommandStr(self.__EncodeSingleObject(modelResource), '/contents', EmptyValue()), transaction);
			#modelResource.contents.clear()
			contents = list(resource.contents)
			pContents = self.__EncodeObjects(contents)
			vContents = ValueParser.PythonToValue(pContents)
			self.DoAsTransaction(CommandParser.UpdateCommandStr(self.__EncodeSingleObject(modelResource), '/contents', EmptyValue() ), transaction);
			self.DoAsTransaction(CommandParser.UpdateCommandStr(self.__EncodeSingleObject(modelResource), '/contents[+]', vContents ), transaction);
			#for content in resource.contents:
			#	modelResource.contents.add(content)
				
			self.resourceObjectLUT[resource] = modelResource
			#read the contents to the resource object because the containment removes it
			for content in modelResource.contents:
				resource.append(content)
		elif(modelResource.type == FileResourceTypesE.TEXT):
			pass
		elif(modelResource.type == FileResourceTypesE.RAW):
			pass
		
		self.__SetResourceLoaded(modelResource, transaction)
		self.__SetResourceClean(modelResource, transaction)
		#self.DoAsTransaction(CommandParser.UpdateCommandStr(self.__EncodeSingleObject(modelResource), '/isDirty', BoolValue(v=False)), transaction);
		#self.DoAsTransaction(CommandParser.UpdateCommandStr(self.__EncodeSingleObject(modelResource), '/isLoaded', BoolValue(v=True)), transaction);
		
	def __SaveResource(self,modelResource,transaction):
		pathHasChanged = False
		if(modelResource.lastPersistentPath and os.path.normpath(modelResource.lastPersistentPath) != os.path.normpath(modelResource.actualPath())):
			pathHasChanged = True
		if(modelResource.type == FileResourceTypesE.MODEL): 
			resource = self.__GetResourceForModelResource(modelResource)
			if(resource and pathHasChanged): #resource was renamed or moved
				resource.uri = URI(os.path.join(self.baseDir,modelResource.actualPath()))
			if(None == resource): #new resource 
				resource = self.resourceSet.create_resource(os.path.join(self.baseDir,modelResource.actualPath()))
				self.resourceObjectLUT[resource] = modelResource
			#update the content
			self.__ConnectAllResources()
			resource.save()
			self.__DisconnectAllResources()
		elif(modelResource.type == FileResourceTypesE.TEXT):
			pass
		elif(modelResource.type == FileResourceTypesE.RAW):
			pass
		self.__SetResourceLoaded(modelResource, transaction)
		self.__SetResourceClean(modelResource, transaction)
		#self.DoAsTransaction(CommandParser.UpdateCommandStr(self.__EncodeSingleObject(modelResource), '/isDirty', BoolValue(v=False)), transaction);
		#self.DoAsTransaction(CommandParser.UpdateCommandStr(self.__EncodeSingleObject(modelResource), '/isLoaded', BoolValue(v=True)), transaction);
		if(pathHasChanged):
			os.remove(os.path.join(self.baseDir,modelResource.lastPersistentPath))
		#modelResource.lastPersistentPath = modelResource.actualPath()
		newPath = modelResource.actualPath()
		self.__SetResourcePersistentPath(modelResource,newPath,transaction)
		
		#self.DoAsTransaction(CommandParser.UpdateCommandStr(self.__EncodeSingleObject(modelResource), '/lastPersistentPath', StringValue(v=modelResource.actualPath())), transaction);
		
	def __GetResourceForModelResource(self,modelResource):
		resource = None
		if(modelResource.lastPersistentPath):
			resourceUri = URI(os.path.join(self.baseDir,modelResource.lastPersistentPath))
			if(resourceUri.normalize() in self.resourceSet.resources): #existing resource
				resource = self.resourceSet.resources[resourceUri.normalize()]
		return resource
	
	def __GetModelResourceForResource(self,resource):
		try:
			return self.resourceObjectLUT[resource]
		except:
			return None
		
	def __ConnectAllResources(self):
		for resourceUri in self.resourceSet.resources:
			resource = self.resourceSet.resources[resourceUri]
			modelResource = self.__GetModelResourceForResource(resource)
			if(modelResource):
				for content in resource.contents:
					resource.remove(content)
				for content in modelResource.contents:
					resource.append(content)
					#modelResource.contents.discard(content)
				#HACK1: elements must be removed from the Resource in order to make save work
				modelResource.contents.clear()
		
	def __DisconnectAllResources(self):
		for resourceUri in self.resourceSet.resources:
			resource = self.resourceSet.resources[resourceUri]
			modelResource = self.__GetModelResourceForResource(resource)
			if(modelResource):
				#HACK2: now read the content
				for content in resource.contents:
					modelResource.contents.add(content)
				#HACK3: content must be re-added to the resource because eResource() wont work otherwise
				for content in modelResource.contents:
					resource.append(content)
				
	
	def __ResourceObjOfEResource(self,eResource):
		resource = None
		resourceUri = URI(eResource.lastPersistentPath)
		if(resourceUri.normalize() in self.resourceSet.resources):
			resource = self.resourceSet.resources(resourceUri.normalize())
		return resource
	
	def __ECloneClass(self,eObject):
		clazz = eObject.eClass
		clone = clazz()
			
		return clone
	
	def __ECloneAttributes(self,eObject):
		clone = self.__ECloneClass(eObject)
		clazz = eObject.eClass
		attributes = clazz.eAllAttributes()
		for attribute in attributes:
			if(attribute.many):
				for attributeValue in eObject.eGet(attribute):
					clone.eGet(attribute).add(attributeValue)
			else:
				clone.eSet(attribute,eObject.eGet(attribute))	
		return clone
	
	def __ECloneDeep(self,eObject,copyReferences=True):
		clone = self.__ECloneAttributes(eObject)
		clazz = eObject.eClass
		references = clazz.eAllReferences()
		for reference in references:
			if(reference.containment):
				if(reference.many):
					for child in eObject.eGet(reference):
						clonedChild = self.__ECloneDeep(child,copyReferences)
						clone.eGet(reference).add(clonedChild)
				else:
					child = eObject.eGet(reference)
					if(child): #only clone non empty single containments
						clonedChild = self.__ECloneDeep(child,copyReferences)
						clone.eSet(reference,clonedChild)
			elif(copyReferences): #non containments
				if(reference.many):
					for refObj in eObject.eGet(reference):
						clone.eGet(reference).add(refObj)
				else:
					clone.eSet(reference,eObject.eGet(reference))
			
		return clone
	
	def __ECloneFull(self,eObject):
		clone = self.__ECloneDeep(eObject,False)
		clazz = eObject.eClass
		self.__ECloneFullReferenceUpdater(eObject,clone,eObject,clone)			
		return clone
	
	def __ECloneFullReferenceUpdater(self,eObject,clonedEObject,root,clone):
		clazz = eObject.eClass
		references = clazz.eAllReferences()
		for reference in references:
			if(reference.containment):
				if(reference.many):
					nChilds = len(eObject.eGet(reference))
					for i in range(nChilds):
						child = eObject.eGet(reference)[i]
						clonedChild = clonedEObject.eGet(reference)[i]
						self.__ECloneFullReferenceUpdater(child, clonedChild, root, clone)
				else:
					child = eObject.eGet(reference)
					if(child): #only clone non empty single containments
						clonedChild = clonedEObject.eGet(reference)
						self.__ECloneFullReferenceUpdater(child, clonedChild, root, clone)
			else: #non containments
				if(reference.many):
					for oldRef in eObject.eGet(reference):
						newRef = self.__ECloneFullFindCorrespondingReference(oldRef,root,clone)
						clonedEObject.eGet(reference).add(newRef)
				else:
					oldRef = eObject.eGet(reference)
					if(oldRef): #null references need no cloning
						newRef = self.__ECloneFullFindCorrespondingReference(oldRef,root,clone)
						clonedEObject.eSet(reference,newRef)
		return clone
	
	def __ECloneFullFindCorrespondingReference(self,oldRef,root,clone):
		newRef = oldRef
		if(self.__ECloneFullIsParentOf(oldRef,root)): #we only need an new reference if the root contains the referred object
			featurePath = self.__ECloneFullGetPathFromTo(root,oldRef)
			newRef = clone
			for p in featurePath: #p is a tuple (featureName, many, index)
				feature = p[0]
				if(p[1]): #many
					index = p[2]
					newRef = newRef.eGet(feature)[index]
				else: #single
					newRef = newRef.eGet(feature)
		return newRef
	
	def __ECloneFullIsParentOf(self,child,possibleParent):
		if(child == possibleParent) :
			return True
		parent = child.eContainer()
		if(parent and isinstance(parent, EObject)):
			return self.__ECloneFullIsParentOf(parent, possibleParent)
		return False #has no parent any more, return false
	
	def __ECloneFullGetPathFromTo(self,anchestor,child):
		featurePath = []
		if(anchestor!=child):
			parent = child.eContainer()
			if(parent and isinstance(parent, EObject)):
				feature = child.eContainmentFeature()
				featureName = feature.name
				many = feature.many
				index = 0
				if(many):
					index = parent.eGet(featureName).index(child)
				featurePath = [(featureName,many,index)]
				featurePath = self.__ECloneFullGetPathFromTo(anchestor,parent) + featurePath
		return featurePath #has no parent any more, return false
			
				
				


from .xgee import DEFAULT_PATH,start
import sys

def main():
    path2apps=DEFAULT_PATH
    if len(sys.argv)>1:
        path2apps=sys.argv[1]
    start(path2apps)     

if __name__ == "__main__":
    main()



from .pyecoresimpleobjectcodec import * 
from .pyecoreidcodec import *
from .pyecoremdbaccessor import *
from eoq2.mdb.pyecore.pyecoresinglefilemdbprovider import *
from .pyecoreworkspacemdbprovider import *



### EXCEPTION DEFINITIONS ###

EOQ_ERROR_TYPE_PARSE = 1000
EOQ_ERROR_TYPE_CONTEXT = 2000
EOQ_ERROR_TYPE_RETRIEVE = 3000
EOQ_ERROR_TYPE_UPDATE = 4000
EOQ_ERROR_TYPE_CREATE = 5000
EOQ_ERROR_TYPE_ID = 6000
EOQ_ERROR_TYPE_VALUE = 7000
EOQ_ERROR_TYPE_TRANSACTION = 8000
EOQ_ERROR_TYPE_ACTION = 8500
EOQ_ERROR_TYPE_GENERAL = 9999



class EoqError(Exception):
    def __init__(self,type,code):
        self.type = type
        self.code = code #all errors will show a specific code

class EoqParseError(EoqError):
    def __init__(self, code, message, query):
        super().__init__(EOQ_ERROR_TYPE_PARSE,code)
        self.message = message
        self.query = query

class EoqContextError(EoqError):
    def __init__(self, code, message):
        super().__init__(EOQ_ERROR_TYPE_CONTEXT,code)
        self.message = message

class EoqRetrieveError(EoqError):
    def __init__(self, code, message):
        super().__init__(EOQ_ERROR_TYPE_RETRIEVE,code)
        self.message = message

class EoqUpdateError(EoqError):
    def __init__(self, code, message, value):
        super().__init__(EOQ_ERROR_TYPE_UPDATE,code)
        self.message = message
        self.value = value

class EoqCreateError(EoqError):
    def __init__(self, code, message, package, clazz):
        super().__init__(EOQ_ERROR_TYPE_CREATE,code)
        self.message = message
        self.package = package
        self.clazz = clazz
        
class EoqIdError(EoqError):
    def __init__(self, code, message, idNb):
        super().__init__(EOQ_ERROR_TYPE_ID,code)
        self.message = message
        self.idNb = idNb
        
class EoqValueError(EoqError):
    def __init__(self, code, message, value):
        super().__init__(EOQ_ERROR_TYPE_VALUE,code)
        self.message = message
        self.value = value
        
class EoqTransactionError(EoqError):
    def __init__(self, code, message):
        super().__init__(EOQ_ERROR_TYPE_TRANSACTION,code)
        self.message = message
        
class EoqActionError(EoqError):
    def __init__(self, code, message):
        super().__init__(EOQ_ERROR_TYPE_ACTION,code)
        self.message = message
        
class EoqGeneralError(EoqError):
    def __init__(self, code, message):
        super().__init__(EOQ_ERROR_TYPE_CREATE,code)
        self.message = message